from typing import Tuple, List, Union, Dict, Any
from builtins import float
import copy
import pandas as pd
from aiphoria.lib.odym.modules.ODYM_Classes import MFAsystem
from .parameters import StockDistributionParameterValueType
from .types import FunctionType, ChangeType


class ObjectBase(object):
    """
    Base class for Process, Flow and Stock.
    Keeps track of row number, validity of read row and virtual state.
    """
    def __init__(self):
        self._id: Union[str, any] = -1
        self._row_number: int = -1
        self._is_valid: bool = False
        self._is_virtual: bool = False

    @property
    def is_valid(self) -> bool:
        """
        Return validity of object.

        :return: Boolean
        """
        return False

    @property
    def id(self) -> Union[str, any]:
        """
        Get ID of the object.

        :return: Object ID (string)
        """
        return self._id

    @id.setter
    def id(self, new_id: str) -> None:
        """
        Set Object ID.

        :param new_id: New ID
        :return: None
        """
        self._id = new_id

    @property
    def row_number(self) -> int:
        """
        Get row number where Object was defined.
        This is valid only for the Flows and Processes defined in settings file.

        :return: Row number where Object was defined
        """
        return self._row_number

    @row_number.setter
    def row_number(self, value: int) -> None:
        """
        Set row number where Object was defined.

        :param value: Target row number
        :return: None
        """
        self._row_number = value

    @property
    def is_virtual(self) -> bool:
        """
        Check if Object is virtual.

        :return: True if Object is virtual, False otherwise.
        """
        return self._is_virtual

    @is_virtual.setter
    def is_virtual(self, value: bool) -> None:
        """
        Set Object virtual state.

        :param value: State
        :return: None
        """
        self._is_virtual = value


class Indicator(object):
    def __init__(self, name: str = None, conversion_factor: float = 1.0, comment: str = None, unit: str = None):
        super().__init__()
        self._name: Union[str, None] = name
        self._conversion_factor: Union[float, None] = conversion_factor
        self._comment: Union[str, None] = comment
        self._unit: Union[str, None] = unit

    @property
    def name(self) -> str:
        """
        Get the indicator name.

        :return: Indicator name (str)
        """
        return self._name

    @name.setter
    def name(self, new_name: str):
        """
        Set the indicator name.

        :param new_name: New indicator name (str)
        """
        self._name = new_name

    @property
    def conversion_factor(self) -> float:
        """
        Get the indicator conversion factor.

        :return: Conversion factor (float)
        """
        return self._conversion_factor

    @conversion_factor.setter
    def conversion_factor(self, new_conversion_factor: float):
        """
        Set the indicator conversion factor.

        :param new_value: New conversion factor (float)
        """
        self._conversion_factor = new_conversion_factor

    @property
    def comment(self) -> str:
        """
        Get indicator comment.

        :return: Indicator comment (str)
        """
        return self._comment

    @comment.setter
    def comment(self, new_comment: str):
        """
        Set the indicator comment.

        :param new_comment: New indicator comment (str)
        """
        self._comment = new_comment

    @property
    def unit(self) -> str:
        """
        Get the indicator unit.

        :return: Indicator unit (str)
        """
        return self._unit

    @unit.setter
    def unit(self, new_unit: str):
        """
        Set the indicator unit.

        :param new_unit: New Indicator unit (str)
        """
        self._unit = new_unit


class Process(ObjectBase):
    """
    aiphoria Process-object:

    Used to store data for Process.
    """
    def __init__(self, params: pd.Series = None, row_number=-1):
        super().__init__()

        self._name = None
        self._location = None
        self._id = None
        self._transformation_stage = None
        self._stock_lifetime = None
        self._stock_lifetime_source = None
        self._stock_distribution_type = None
        self._stock_distribution_params = None
        self._wood_content = None
        self._wood_content_source = None
        self._density = None
        self._density_source = None
        self._modelling_status = None
        self._comment = None
        self._row_number = -1
        self._depth = -1
        self._position_x = None
        self._position_y = None
        self._label_in_graph = None

        # Leave instance to default state if params is not provided
        if params is None:
            return

        # Skip totally empty row
        if params.isna().all():
            return

        self._name = params.iloc[0]
        self._location = params.iloc[1]
        self._id = params.iloc[2]
        self._transformation_stage = params.iloc[3]

        # Parse stock lifetime, default to zero if None
        self._stock_lifetime = self._parse_stock_lifetime(params.iloc[4], row_number)

        self._stock_lifetime_source = params.iloc[5]
        self._stock_distribution_type = params.iloc[6]
        self._stock_distribution_params = params.iloc[7]

        # Parse stock distribution parameters
        # NOTE: Event invalid key-value -pairs are stored to _stock_distribution_params after parsin
        # and those are checked in datachecker
        self._parse_and_set_distribution_params(params.iloc[7])

        self._wood_content = params.iloc[8]
        self._wood_content_source = params.iloc[9]
        self._density = params.iloc[10]
        self._density_source = params.iloc[11]
        self._modelling_status = params.iloc[12]
        self._comment = params.iloc[13]
        self._position_x = params.iloc[14]
        self._position_y = params.iloc[15]
        self._label_in_graph = params.iloc[16]
        self._row_number = row_number
        self._meta = {}

    def __str__(self) -> str:
        s = "Process '{}': Lifetime: {}".format(self.id, self.stock_lifetime)
        return s

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other) -> bool:
        if not isinstance(other, Process):
            return NotImplemented

        return self.id == other.id

    def is_valid(self) -> bool:
        is_valid = True
        is_valid = is_valid and self.name is not None
        is_valid = is_valid and self.location is not None
        is_valid = is_valid and self.id is not None
        return is_valid

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value

    @name.setter
    def name(self, value) -> None:
        self._name = value

    @property
    def location(self) -> str:
        return self._location

    @location.setter
    def location(self, value: str):
        self._location = value

    @property
    def transformation_stage(self) -> str:
        return self._transformation_stage

    @transformation_stage.setter
    def transformation_stage(self, value: str):
        self._transformation_stage = value

    @property
    def stock_lifetime(self) -> int:
        return self._stock_lifetime

    @stock_lifetime.setter
    def stock_lifetime(self, value: int) -> None:
        self._stock_lifetime = value

    @property
    def stock_lifetime_source(self) -> str:
        return self._stock_lifetime_source

    @stock_lifetime_source.setter
    def stock_lifetime_source(self, value: str):
        self._stock_lifetime_source = value

    @property
    def stock_distribution_type(self) -> str:
        return self._stock_distribution_type

    @stock_distribution_type.setter
    def stock_distribution_type(self, value: str):
        self._stock_distribution_type = value

    @property
    def stock_distribution_params(self) -> str:
        return self._stock_distribution_params

    @stock_distribution_params.setter
    def stock_distribution_params(self, value: str):
        self._stock_distribution_params = value

    @property
    def wood_content(self) -> float:
        return self._wood_content

    @wood_content.setter
    def wood_content(self, value: float):
        self._wood_content = value

    @property
    def wood_content_source(self) -> str:
        return self._wood_content_source

    @wood_content_source.setter
    def wood_content_source(self, value: str):
        self._wood_content_source = value

    @property
    def density(self) -> float:
        return self._density

    @density.setter
    def density(self, value: float):
        self._density = value

    @property
    def density_source(self) -> str:
        return self._density_source

    @density_source.setter
    def density_source(self, value: str):
        self._density_source = value

    @property
    def modelling_status(self) -> str:
        return self._modelling_status

    @modelling_status.setter
    def modelling_status(self, value: str):
        self._modelling_status = value

    @property
    def comment(self) -> str:
        return self._comment

    @comment.setter
    def comment(self, value: str):
        self._comment = value

    @property
    def depth(self) -> int:
        return self._depth

    @depth.setter
    def depth(self, value: int):
        self._depth = value

    @property
    def position_x(self) -> float:
        return self._position_x

    @position_x.setter
    def position_x(self, value: float):
        self._position_x = value

    @property
    def position_y(self) -> float:
        return self._position_y

    @position_y.setter
    def position_y(self, value: float):
        self._position_y = value

    @property
    def label_in_graph(self) -> str:
        return self._label_in_graph

    @label_in_graph.setter
    def label_in_graph(self, value: str):
        self._label_in_graph = value

    @property
    def meta(self) -> Dict[Any, Any]:
        return self._meta

    @meta.setter
    def meta(self, value: Dict[Any, Any]) -> None:
        self._meta = value

    def _parse_stock_lifetime(self, s: str, row_number: int = -1):
        """
        Parse stock lifetime from string.

        :param s: String
        :return: Stock lifetime (int)
        """
        lifetime = 0
        if s is None:
            return lifetime

        try:
            lifetime = int(s)
        except (ValueError, TypeError) as ex:
            raise Exception("Stock lifetime must be number (row {})".format(row_number))

        return lifetime

    def _parse_and_set_distribution_params(self, s: str):
        """
        Parse keys from string for distribution parameters.
        """

        try:
            # Check if cell contains only value
            self._stock_distribution_params = float(s)
            return
        except (ValueError, TypeError):
            if s is None:
                return

        params = {}
        has_multiple_params = s.find(',') > 0
        if not has_multiple_params:
            # Single parameter
            entry = s
            k = None
            v = None
            if entry.find("=") >= 0:
                # Has key=value
                k, v = entry.split("=")
                k = k.strip()
                v = v.strip()
            else:
                # Only key, no value
                k = entry.strip()
                v = None

            # Convert to target value type if definition exists
            value_type = StockDistributionParameterValueType[k]
            if value_type is not None:
                v = value_type(v)

            params[k] = v

        else:
            # Multiple parameters, separated by ','
            for entry in s.split(","):
                k = None
                v = None
                if entry.find("=") >= 0:
                    k, v = entry.split("=")
                    k = k.strip()
                    v = v.strip()
                else:
                    k = entry.strip()
                    v = None

                # Convert to target value type if definition exists
                value_type = StockDistributionParameterValueType[k]
                if value_type is not None:
                    v = value_type(v)

                params[k] = v

        self._stock_distribution_params = params


class Flow(ObjectBase):
    """
    aiphoria Flow-object:

    Used to store data for Flow.
    """
    def __init__(self, params: pd.Series = None, row_number=-1):
        super().__init__()

        self._source_process = None
        self._source_process_transformation_stage = None
        self._source_process_location = None
        self._target_process = None
        self._target_process_transformation_stage = None
        self._target_process_location = None
        self._source_process_id = None
        self._target_process_id = None
        self._value = None
        self._unit = None
        self._year = None
        self._data_source = None
        self._data_source_comment = None
        self._comment = None

        # Evaluated per timestep
        self._is_evaluated = False
        self._evaluated_share = 0.0
        self._evaluated_value = 0.0

        # Indicator name to Indicator
        self._indicator_name_to_indicator = {}
        self._indicator_name_to_evaluated_value = {}

        # Flow prioritization
        self._is_prioritized = False

        if params is None:
            return

        # Skip totally empty row
        if params.isna().all():
            return

        self._source_process = params.iloc[0]
        self._source_process_transformation_stage = params.iloc[1]
        self._source_process_location = params.iloc[2]
        self._target_process = params.iloc[3]
        self._target_process_transformation_stage = params.iloc[4]
        self._target_process_location = params.iloc[5]
        self._source_process_id = params.iloc[6]
        self._target_process_id = params.iloc[7]
        self._value = params.iloc[8]
        self._unit = params.iloc[9]
        self._year = int(params.iloc[10])
        self._data_source = params.iloc[11]
        self._data_source_comment = params.iloc[12]

        # Rest of the elements except last element are indicators
        # There should be even number of indicators because each indicator has value and comment
        first_indicator_index = 13
        indicators = params[first_indicator_index:]
        if len(indicators) % 2:
            s = "Not even number of indicator columns in settings file.\n"
            s += "Each indicator needs two columns (value and comment) in this order."
            raise Exception(s)

        # Build indicator name to Indicator mappings
        for i in range(0, len(indicators), 2):
            indicator_name = indicators.index[i]
            conversion_factor = indicators.iloc[i]
            comment = indicators.iloc[i+1]

            # Strip substring inside characters '(' and  ')'
            # and use that as a unit
            indicator_unit = ""
            start_index = indicator_name.find("(")
            end_index = indicator_name.find(")")
            if start_index >= 0 and end_index >= 0:
                unit_name = indicator_name[start_index:end_index + 1]
                indicator_name = indicator_name.replace(unit_name, '').strip()
                indicator_unit = unit_name[1:-1].strip()

            # NOTE: Set indicator conversion factor to 0.0
            # if not defined in the settings file
            if conversion_factor is None:
                conversion_factor = 0.0

            new_indicator = Indicator(indicator_name, conversion_factor, comment, indicator_unit)
            self._indicator_name_to_indicator[indicator_name] = new_indicator
            self._indicator_name_to_evaluated_value[indicator_name] = 0.0

        self._row_number = row_number  # Track Excel file row number

    def __str__(self):
        s = "Flow '{}' -> '{}': Value={}, Unit={}, " \
            "is_evaluated={}, evaluated_share={}, evaluated_value={}, " \
            "year={}, is_virtual={}".format(
                self.source_process_id, self.target_process_id, self.value, self.unit,
                self.is_evaluated, self.evaluated_share, self.evaluated_value, self.year,
                self.is_virtual)
        return s

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other):
        if not isinstance(other, Flow):
            return NotImplemented

        return self.id == other.id

    @staticmethod
    def make_flow_id(source_process_id: str, target_process_id: str) -> str:
        """
        Make Flow ID from source Process ID and target Process ID.

        :param source_process_id: Source Process ID (string)
        :param target_process_id: Target Process ID (string)
        :return: Flow ID (string)
        """
        return "{} {}".format(source_process_id, target_process_id)

    @property
    def id(self) -> str:
        """
        Returns Flow ID.

        :return: Flow ID (string)
        """
        return Flow.make_flow_id(self.source_process_id, self.target_process_id)

    def is_valid(self):
        is_valid = True
        is_valid = is_valid and self.value is not None
        is_valid = is_valid and (self.source_process is not None)
        is_valid = is_valid and (self.target_process is not None)
        is_valid = is_valid and (self.source_process_id is not None)
        is_valid = is_valid and (self.target_process_id is not None)
        return is_valid

    @property
    def is_unit_absolute_value(self):
        # Default to absolute value if unit is missing
        unit_str = self.unit
        if unit_str is None:
            return True

        unit_str = unit_str.strip()
        if unit_str == "%":
            return False

        return True

    @property
    def source_process(self) -> str:
        """
        Get source Process name.

        :return: Source Process name (str)
        """
        return self._source_process

    @property
    def source_process_transformation_stage(self) -> str:
        return self._source_process_transformation_stage

    @property
    def source_process_location(self) -> str:
        return self._source_process_location

    @property
    def target_process(self) -> str:
        """
        Get target Process name.
        :return: Target Process name (str)
        """
        return self._target_process

    @property
    def target_process_transformation_stage(self) -> str:
        return self._target_process_transformation_stage

    @property
    def target_process_location(self) -> str:
        return self._target_process_location

    @property
    def source_process_id(self) -> str:
        """
        Get source Process ID.
        :return: Source Process ID (str)
        """
        return self._source_process_id

    @source_process_id.setter
    def source_process_id(self, source_process_id: str):
        """
        Set source Process ID.

        :param source_process_id: Source Process ID (str)
        """
        self._source_process_id = source_process_id

    @property
    def target_process_id(self) -> str:
        """
        Get target Process ID.

        :return: Target Process ID (str)
        """
        return self._target_process_id

    @target_process_id.setter
    def target_process_id(self, target_process_id: str):
        """
        Set target Process ID.

        :param target_process_id: New target Process ID
        """
        self._target_process_id = target_process_id

    # Original value from Excel row
    @property
    def value(self) -> float:
        """
        Get original baseline value.

        :return: Original baseline value (float)
        """
        return self._value

    @value.setter
    def value(self, value: float):
        """
        Set original base value.

        :param value: New original base value (float)
        """
        self._value = value

    @property
    def unit(self) -> str:
        return self._unit

    @unit.setter
    def unit(self, unit: str):
        self._unit = unit

    @property
    def year(self) -> int:
        return self._year

    @year.setter
    def year(self, value: int):
        self._year = value

    @property
    def data_source(self) -> str:
        return self._data_source

    @property
    def data_source_comment(self) -> str:
        return self._data_source_comment

    @property
    def comment(self) -> str:
        return self._comment

    @property
    def is_evaluated(self) -> bool:
        """
        Get flow evaluated state.

        :return: True if Flow is evaluated, False otherwise.
        """
        return self._is_evaluated

    @is_evaluated.setter
    def is_evaluated(self, value: bool):
        """
        Set flow evaluated state.

        :param value: New evaluated state (bool)
        """
        self._is_evaluated = value

    @property
    def evaluated_value(self) -> float:
        """
        Get evaluated base value.

        :return: Evaluated base value (float)
        """
        return self._evaluated_value

    @evaluated_value.setter
    def evaluated_value(self, value: float):
        """
        Set evaluated base value.

        :param value: New evaluated base value (float)
        """
        self._evaluated_value = value

    @property
    def evaluated_share(self) -> float:
        return self._evaluated_share

    @evaluated_share.setter
    def evaluated_share(self, value: float):
        self._evaluated_share = value

    @property
    def is_prioritized(self) -> bool:
        return self._is_prioritized

    @is_prioritized.setter
    def is_prioritized(self, is_prioritized: bool):
        self._is_prioritized = is_prioritized

    @property
    def indicator_name_to_indicator(self) -> Dict[str, Indicator]:
        return self._indicator_name_to_indicator

    @property
    def indicator_name_to_evaluated_value(self) -> Dict[str, float]:
        return self._indicator_name_to_evaluated_value

    def get_indicator_names(self) -> List[str]:
        """
        Get list of Indicator names (including baseline indicator name).

        :return: List of Indicator names
        """
        return list(self._indicator_name_to_indicator.keys())

    def get_indicator_units(self) -> List[str]:
        """
        Get list of Indicator unit names (including baseline indicator unit).

        :return: List of Indicator unit names
        """
        indicator_units = []
        for indicator_name, indicator in self._indicator_name_to_indicator.items():
            indicator_units.append(indicator.unit)

        return indicator_units

    def get_indicator_conversion_factor(self, indicator_name: str) -> float:
        """

        :param indicator_name:
        :return:
        """
        return self._indicator_name_to_indicator[indicator_name].conversion_factor

    def get_evaluated_value_for_indicator(self, indicator_name: str) -> float:
        """
        Get evaluated value for Indicator.

        :param indicator_name: Target Indicator name (str)
        :return: Evaluated value for Indicator (float)
        """
        return self._indicator_name_to_evaluated_value[indicator_name]

    def set_evaluated_value_for_indicator(self, indicator_name: str, value: float):
        """
        Set evaluated value for Indicator.

        :param indicator_name: Target Indicator name (str)
        :param value: New evaluated value for Indicator (float)
        """
        self._indicator_name_to_evaluated_value[indicator_name] = value

    def evaluate_indicator_values_from_baseline_value(self):
        """
        Evaluated indicator evaluated value from baseline value.
        """
        for indicator_name, indicator in self._indicator_name_to_indicator.items():
            evaluated_value = self.evaluated_value * indicator.conversion_factor
            self.set_evaluated_value_for_indicator(indicator_name, evaluated_value)

    def get_all_evaluated_values(self) -> List[float]:
        """
        Get list of all evaluated values.
        First index is always the evaluated baseline value.
        Other indices are evaluated indicator values

        :return: List of evaluated values (list of float)
        """

        return [self.evaluated_value] + [value for name, value in self.indicator_name_to_evaluated_value.items()]


# Stock is created for each process that has lifetime
class Stock(ObjectBase):
    """
    aiphoria Stock-object:

    Used to store data for Stock.
    """
    def __init__(self, params: Process = None, row_number=-1):
        super().__init__()
        self._process = None
        self._id = -1

        if params is None:
            return

        self._process = params
        self._id = params.id
        self._row_number = row_number

    def __str__(self):
        if not self.is_valid():
            return "Stock: no process"

        s = "Stock: Process='{}', lifetime={}".format(self.id, self.stock_lifetime)
        return s

    def is_valid(self):
        if not self._process:
            return False

        return True

    def __hash__(self):
        return hash(self._process.id)

    def __eq__(self, other):
        return self.id == other.id

    @property
    def name(self):
        return self._process.name

    @property
    def stock_lifetime(self):
        return self._process.stock_lifetime

    @property
    def stock_distribution_type(self):
        return self._process.stock_distribution_type

    @property
    def stock_distribution_params(self):
        return self._process.stock_distribution_params


class FlowModifier(ObjectBase):
    """
    aiphoria FlowModifier-object:

    Used to store data for scenario related flow modifications.
    """
    def __init__(self, params: pd.Series = None):
        super().__init__()

        self._scenario_name: str = ""
        self._source_process_id: str = ""
        self._target_process_id: str = ""
        self._change_in_value: Union[float, None] = None
        self._target_value: Union[float, None] = None
        self._change_type: str = ""
        self._start_year: int = 0
        self._end_year: int = 0
        self._function_type: str = ""
        self._apply_to_targets: bool = True
        self._opposite_target_process_ids = []

        if params is None:
            # Invalid: no parameters
            return

        if all(not elem for elem in params):
            # Invalid: all parameters None
            return

        # Alias parameters to more readable form
        param_scenario_name = params.iloc[0]
        param_source_process_id = params.iloc[1]
        param_target_process_id = params.iloc[2]
        param_change_in_value = params.iloc[3]
        param_target_value = params.iloc[4]
        param_change_type = params.iloc[5]
        param_start_year = params.iloc[6]
        param_end_year = params.iloc[7]
        param_function_type = params.iloc[8]
        param_apply_to_targets = params.iloc[9]

        self._scenario_name = self._parse_as(param_scenario_name, str)[0]
        self._source_process_id = self._parse_as(param_source_process_id, str)[0]
        self._target_process_id = self._parse_as(param_target_process_id, str)[0]

        # This is the delta change of the value and means that it's error
        # if target flow has ABS type and the 'change in value' is REL
        # NOTE: Either of self._change_in_value of self._target_value must be defined
        if param_change_in_value is not None:
            self._change_in_value = self._parse_as(param_change_in_value, float)[0]

        if param_target_value is not None:
            self._target_value = self._parse_as(param_target_value, float)[0]

        # Change type
        # NOTE: Convert change type to valid ChangeType enum if found. Otherwise use the value from parameter.
        self._change_type = param_change_type
        for change_type in ChangeType:
            if self._parse_as(self._change_type, str)[0].lower() == change_type.lower():
                self._change_type = change_type

        self._start_year = self._parse_as(param_start_year, int)[0]
        self._end_year = self._parse_as(param_end_year, int)[0]

        # Function type
        # NOTE: Convert function type to valid FunctionType enum if found. Otherwise use the value from parameter.
        self._function_type = param_function_type
        for function_type in FunctionType:
            if self._parse_as(self._function_type, str)[0].lower() == function_type:
                self._function_type = function_type

        # Apply to siblings
        if param_apply_to_targets is not None:
            self._apply_to_targets = self._parse_as(param_apply_to_targets, bool)[0]

        # Check how many target nodes with opposite effect there is
        for process_id in list(params[10:]):
            if process_id is not None:
                self._opposite_target_process_ids.append(process_id)

    def __str__(self):
        s = "Flow modifier: scenario_name='{}', source_process_id='{}', target_process_id='{}', change_in_value='{}', " \
            "target_value='{}', change_type='{}', start_year='{}', end_year='{}', function_type='{}'".format(
            self.scenario_name,
            self.source_process_id,
            self.target_process_id,
            self.change_in_value,
            self.target_value,
            self.change_type,
            self.start_year,
            self.end_year,
            self.function_type,
        )
        return s

    def is_valid(self) -> bool:
        return True

    @property
    def use_change_in_value(self) -> bool:
        return self.change_in_value is not None

    @property
    def use_target_value(self) -> bool:
        return self.target_value is not None

    @property
    def is_change_type_value(self) -> bool:
        """
        Check if change type is value type (= absolute change in flow value or in flow share)
        :return: True if change type is value type, False otherwise.
        """
        return self.change_type == ChangeType.Value

    @property
    def is_change_type_proportional(self) -> bool:
        """
        Is change type proportional (= relative change in flow value or in flow share)
        :return: True if change type is proportional, False otherwise.
        """
        return self.change_type == ChangeType.Proportional

    @property
    def has_target(self) -> bool:
        """
        Does flow modifier target any flow?
        :return: Bool
        """
        return self.target_process_id != ""

    @property
    def has_opposite_targets(self) -> bool:
        """
        Does flow modifier have opposite targets?
        :return: Bool
        """
        return len(self.opposite_target_process_ids) > 0


    @staticmethod
    def _parse_as(val: any, target_type: any) -> (bool, any):
        """
        Parse variable as type.
        Returns tuple (bool, value).
        If parsing fails then value is the original val.

        :param val: Variable
        :param type: Target type
        :return: Tuple (bool, value)
        """
        ok = True
        result = val
        try:
            result = target_type(val)
        except ValueError as ex:
            ok = False

        return result, ok

    @property
    def scenario_name(self) -> str:
        return self._scenario_name

    @property
    def source_process_id(self) -> str:
        return self._source_process_id

    @property
    def target_process_id(self) -> str:
        return self._target_process_id

    @property
    def target_flow_id(self) -> str:
        return Flow.make_flow_id(self.source_process_id, self.target_process_id)

    @property
    def change_in_value(self) -> float:
        return self._change_in_value

    @property
    def target_value(self) -> float:
        return self._target_value

    @property
    def change_type(self) -> str:
        return self._change_type

    @property
    def start_year(self) -> int:
        return self._start_year

    @property
    def end_year(self) -> int:
        return self._end_year

    @property
    def function_type(self) -> Union[FunctionType, str]:
        return self._function_type

    @property
    def apply_to_targets(self) -> bool:
        return self._apply_to_targets

    @property
    def opposite_target_process_ids(self) -> List[str]:
        return self._opposite_target_process_ids

    def get_year_range(self) -> List[int]:
        """
        Get list of years FlowModifier is used.

        :return: List of years (integers)
        """
        return [year for year in range(self.start_year, self.end_year + 1)]

    def get_opposite_target_flow_ids(self) -> List[str]:
        """
        Get list of opposite target flow IDs.

        :return: List of opposite target flow IDs.
        """
        opposite_flow_ids = []
        for opposite_process_id in self.opposite_target_process_ids:
            flow_id = Flow.make_flow_id(self.source_process_id, opposite_process_id)
            opposite_flow_ids.append(flow_id)
        return opposite_flow_ids


class ScenarioData(object):
    """
    Data class for holding Scenario data.

    DataChecker builds ScenarioData-object that can be used for FlowSolver.
    """

    def __init__(self,
                 years: List[int] = None,
                 year_to_process_id_to_process: Dict[int, Dict[str, Process]] = None,
                 year_to_process_id_to_flow_ids: Dict[int, Dict[str, Dict[str, List[str]]]] = None,
                 year_to_flow_id_to_flow: Dict[int, Dict[str, Flow]] = None,
                 stocks: List[Stock] = None,
                 process_id_to_stock: Dict[str, Stock] = None,
                 unique_process_id_to_process: Dict[str, Process] = None,
                 unique_flow_id_to_flow: Dict[str, Flow] = None,
                 use_virtual_flows: bool = True,
                 virtual_flows_epsilon: float = 0.1,
                 baseline_value_name: str = "Baseline",
                 baseline_unit_name: str = "Baseline unit",
                 indicator_name_to_indicator: Dict[str, Indicator] = None
                 ):

        if years is None:
            years = []

        if year_to_process_id_to_process is None:
            year_to_process_id_to_process = {}

        if year_to_process_id_to_flow_ids is None:
            year_to_process_id_to_flow_ids = {}

        if year_to_flow_id_to_flow is None:
            year_to_flow_id_to_flow = {}

        if stocks is None:
            stocks = []

        if process_id_to_stock is None:
            process_id_to_stock = {}

        if unique_process_id_to_process is None:
            unique_process_id_to_process = {}

        if unique_flow_id_to_flow is None:
            unique_flow_id_to_flow = {}

        if indicator_name_to_indicator is None:
            indicator_name_to_indicator = {}

        self._year_to_flow_id_to_flow = year_to_flow_id_to_flow
        self._year_to_process_id_to_process = year_to_process_id_to_process
        self._year_to_process_id_to_flow_ids = year_to_process_id_to_flow_ids

        self._years = years
        self._year_start = 0
        self._year_end = 0

        if self._years:
            self._year_start = min(self._years)
            self._year_end = max(self._years)

        self._process_id_to_stock = process_id_to_stock
        self._stocks = stocks
        self._unique_process_id_to_process = unique_process_id_to_process
        self._unique_flow_id_to_flow = unique_flow_id_to_flow
        self._use_virtual_flows = use_virtual_flows
        self._virtual_flows_epsilon = virtual_flows_epsilon
        self._baseline_value_name = baseline_value_name
        self._baseline_unit_name = baseline_unit_name
        self._indicator_name_to_indicator = indicator_name_to_indicator

    @property
    def years(self) -> List[int]:
        """
        Get list of years
        :return: List of years
        """
        return self._years

    @property
    def year_to_process_id_to_process(self) -> Dict[int, Dict[str, Process]]:
        """
        Get year to Process ID to Process mappings.

        :return: Dictionary (Year -> Process ID -> Process)
        """
        return self._year_to_process_id_to_process

    @property
    def year_to_process_id_to_flow_ids(self) -> Dict[int, Dict[str, Dict[str, List[str]]]]:
        """
        Get year to Process ID to In/Out to -> List of Flow ID mappings.

        :return: Dictionary (Year -> Process ID -> Dictionary(keys "in", "out") -> List of Flow IDS)
        """

        return self._year_to_process_id_to_flow_ids

    @property
    def year_to_flow_id_to_flow(self) -> Dict[int, Dict[str, Flow]]:
        """
        Get year to Flow ID to Flow mappings.

        :return: Dictionary (Year -> Flow ID -> Flow)
        """
        return self._year_to_flow_id_to_flow

    @property
    def stocks(self) -> List[Stock]:
        """
        Get list of Stocks
        :return: List of Stocks
        """
        return self._stocks

    @property
    def process_id_to_stock(self) -> Dict[str, Stock]:
        """
        Get mapping of Process ID to Stock
        :return: Dictionary
        """
        return self._process_id_to_stock

    @property
    def unique_process_id_to_process(self) -> Dict[str, Process]:
        """
        Get mapping of unique Process ID to Process
        :return: Dictionary
        """
        return self._unique_process_id_to_process

    @property
    def unique_flow_id_to_flow(self) -> Dict[str, Flow]:
        """
        Get mapping of unique Flow ID to Flows
        :return: Dictionary
        """
        return self._unique_flow_id_to_flow

    @property
    def use_virtual_flows(self) -> bool:
        """
        Get boolean flag if using virtual flows
        :return: bool
        """
        return self._use_virtual_flows

    @property
    def virtual_flows_epsilon(self) -> float:
        """
        Get maximum allowed difference between input and output flows before creating virtual flows.
        This is only used if using the virtual flows.
        :return: Float
        """
        return self._virtual_flows_epsilon

    @property
    def start_year(self) -> int:
        """
        Get starting year
        :return: Starting year (int)
        """
        return self._year_start

    @property
    def end_year(self) -> int:
        """
        Get ending year
        Ending year is included in simulation.

        :return: Ending year (int)
        """
        return self._year_end

    @property
    def baseline_value_name(self) -> str:
        """
        Get baseline value name (e.g. "Solid wood equivalent")

        :return: Baseline value name (str)
        """
        return self._baseline_value_name

    @baseline_value_name.setter
    def baseline_value_name(self, new_name: str):
        """
        Set new baseline value name.

        :param new_name: New baseline value name (str)
        """
        self._baseline_value_name = new_name

    @property
    def baseline_unit_name(self) -> str:
        """
        Get baseline unit name (e.g. "Mm3")

        :return: Baseline unit name (str)
        """
        return self._baseline_unit_name

    @baseline_unit_name.setter
    def baseline_unit_name(self, new_name: str):
        """
        Set new baseline unt name.

        :param new_name: New baseline name (str)
        """
        self._baseline_unit_name = new_name

    @property
    def indicator_name_to_indicator(self) -> Dict[str, Indicator]:
        """
        Get dictionary of Indicator name to Indicator.

        :return: Dictionary (indicator name (str), Indicator)
        """
        return self._indicator_name_to_indicator

    @indicator_name_to_indicator.setter
    def indicator_name_to_indicator(self, new_indicator_name_to_indicator: Dict[str, Indicator]):
        """
        Set new Indicator name to Indicator mapping.

        :param new_indicator_name_to_indicator: New Indicator name to Indicator dictionary
        """
        self._indicator_name_to_indicator = new_indicator_name_to_indicator


class ScenarioDefinition(object):
    """
    ScenarioDefinition is wrapper object that contains scenario name and all flow modifiers that are applied
    for the Scenario.

    Actual building of Scenarios happens inside DataChecker.build_scenarios()
    """
    def __init__(self, name: str = None, flow_modifiers: List[FlowModifier] = None):
        if name is None:
            name = "Baseline scenario"

        if flow_modifiers is None:
            flow_modifiers = []

        self._name = name
        self._flow_modifiers = flow_modifiers

    @property
    def name(self) -> str:
        """
        Get scenario name.

        :return: Scenario name
        """
        return self._name

    @property
    def flow_modifiers(self) -> List[FlowModifier]:
        """
        Get list of FlowModifiers.
        These are the rules that are applied to Scenario.

        :return: List of FlowModifier-objects
        """
        return self._flow_modifiers


class Scenario(object):
    """
    Scenario is wrapper object that contains scenario name and all flow modifiers that are
    happening in the scenario
    """

    def __init__(self, definition: ScenarioDefinition = None, data: ScenarioData = None, model_params=None):
        if definition is None:
            definition = ScenarioDefinition()

        if data is None:
            data = ScenarioData()

        if model_params is None:
            model_params = {}

        self._scenario_definition = definition
        self._scenario_data = data
        self._flow_solver = None
        self._odym_data = None
        self._model_params = model_params
        self._mfa_system = None

    @property
    def name(self) -> str:
        return self._scenario_definition.name

    @property
    def scenario_definition(self) -> ScenarioDefinition:
        return self._scenario_definition

    @property
    def scenario_data(self) -> ScenarioData:
        return self._scenario_data

    # Flow solver
    @property
    def flow_solver(self):
        """
        Get FlowSolver that is assigned to Scenario.
        :return: FlowSolver (FlowSolver)
        """
        return self._flow_solver

    @flow_solver.setter
    def flow_solver(self, flow_solver):
        self._flow_solver = flow_solver

    @property
    def model_params(self) -> Dict[str, Any]:
        return self._model_params

    @property
    def mfa_system(self) -> MFAsystem:
        """
        Get stored ODYM MFA system.
        :return: MFAsystem-object
        """
        return self._mfa_system

    @mfa_system.setter
    def mfa_system(self, mfa_system) -> None:
        """
        Set new MFAsystem
        :param mfa_system: Target MFAsystem-object
        """
        self._mfa_system = mfa_system

    def copy_from_baseline_scenario_data(self, scenario_data: ScenarioData):
        """
        Copy ScenarioData from baseline Scenario.
        Data is deep copied and is not referencing to original data anymore.

        :param scenario_data: ScenarioData from baseline FlowSolver.
        """
        self._scenario_data = copy.deepcopy(scenario_data)


class Color(ObjectBase):
    def __init__(self, params: Union[List, pd.Series] = None, row_number=-1):
        super().__init__()
        self._name: str = ""
        self._value: str = ""
        self.row_number = row_number

        if params is None:
            return

        # Handle processing list and pd.Series differently
        name_val = ""
        value_val = ""
        if isinstance(params, list):
            name_val = str(params[0])
            value_val = str(params[1])

        if isinstance(params, pd.Series):
            name_val = str(params.iloc[0])
            value_val = str(params.iloc[1])

        self.name = name_val
        self.value = value_val

    def __str__(self) -> str:
        """
        Returns the string presentation of the Color in format:
        #rrggbb
        where
            rr = red component
            gg = green component
            bb = blue component

        :return: Color as hexadecimal string, prefixed with character '#'
        """
        return self.value.lower()

    def is_valid(self) -> bool:
        return not self.name and self.value

    @property
    def name(self) -> str:
        """
        Get color name.

        :return: Color name (str)
        """
        return self._name

    @name.setter
    def name(self, new_name: str):
        """
        Set color name.

        :param new_name: New color name
        """
        self._name = new_name

    @property
    def value(self) -> str:
        """
        Color value (hexadecimal, e.g. #AABBCC) prefixed with the character '#'
        """
        return self._value

    @value.setter
    def value(self, new_value: str):
        """
        Set new color value (hexadecimal).
        New color must be prefixed with the character '#'.

        :param new_value: New color value (hexadecimal)
        """
        self._value = new_value

    def get_red_as_float(self) -> float:
        """
        Get red component value in range [0, 1]
        :return: Red value (float)
        """
        return self._hex_to_normalized_float(self.value[1:][0:2])

    def get_green_as_float(self) -> float:
        """
        Get green component value in range [0, 1]
        :return: Green value (float)
        """
        return self._hex_to_normalized_float(self.value[1:][2:4])

    def get_blue_as_float(self) -> float:
        """
        Get blue component value in range [0, 1]
        :return: Blue value (float)
        """
        return self._hex_to_normalized_float(self.value[1:][4:6])

    def _hex_to_normalized_float(self, hex_value: str) -> float:
        return int(hex_value, 16) / 255.0


class ProcessEntry(object):
    """
    Internal storage class for Process entry (process, inflows, and outflows).
    Used when storing Process data in DataFrames.
    """

    KEY_IN: str = "in"
    KEY_OUT: str = "out"

    def __init__(self, process: Process = None):
        """
        Initialize ProcessEntry.
        Makes deep copy of target Process.

        :param process: Target Process
        """
        self._process = process
        self._flows = {self.KEY_IN: {}, self.KEY_OUT: {}}

    @property
    def process(self) -> Process:
        return self._process

    @property
    def flows(self) -> Dict[str, Dict[str, Flow]]:
        return self._flows

    @flows.setter
    def flows(self, flows: Dict[str, Dict[str, Flow]]):
        self._flows = flows

    @property
    def inflows(self) -> List[Flow]:
        return list(self._flows[self.KEY_IN].values())

    @inflows.setter
    def inflows(self, flows: List[Flow]):
        self._flows[self.KEY_IN] = {flow.id: flow for flow in flows}

    @property
    def inflows_as_dict(self) -> Dict[str, Flow]:
        return self._flows[self.KEY_IN]

    @property
    def outflows(self) -> List[Flow]:
        return list(self._flows[self.KEY_OUT].values())

    @outflows.setter
    def outflows(self, flows: List[Flow]):
        self._flows[self.KEY_OUT] = {flow.id: flow for flow in flows}

    def outflows_as_dict(self) -> Dict[str, Flow]:
        return self._flows[self.KEY_OUT]

    def add_inflow(self, flow: Flow):
        self._flows[self.KEY_IN][flow.id] = flow

    def add_outflow(self, flow: Flow):
        self._flows[self.KEY_OUT][flow.id] = flow

    def remove_inflow(self, flow_id: str):
        """
        Remove inflow by Flow ID.
        Raises Exception if Flow ID is not found in inflows.

        :param flow_id: Target Flow ID
        :raises Exception If Flow ID is not found
        """
        removed_flow_id = self._flows[self.KEY_IN].pop(flow_id, None)
        if not removed_flow_id:
            raise Exception("No flow_id {} in inflows".format(flow_id))

    def remove_outflow(self, flow_id: str):
        """
        Remove outflow by Flow ID.
        Raises Exception if Flow ID is not found in outflows.

        :param flow_id: Target Flow ID
        :raises Exception If Flow ID is not found
        """
        removed_flow_id = self._flows[self.KEY_OUT].pop(flow_id, None)
        if not removed_flow_id:
            raise Exception("No flow_id {} in outflows".format(flow_id))



import os
import pickle
import shutil
from typing import Union, List
from . import logger
from .logger import log, start_log_perf, stop_log_perf, clear_log_perf, show_log_perf_summary
from .datachecker import DataChecker
from .dataprovider import DataProvider
from .datastructures import Scenario
from .flowsolver import FlowSolver
from .parameters import ParameterName
from .utils import show_exception_errors, show_model_parameters, build_mfa_system_for_scenario

# Globals
global_path_to_cache = ""
global_use_cache = False
global_use_timing = False
global_clear_cache = False


def init_builder(path_to_cache: str,
                 use_cache: bool = False,
                 use_timing: bool = False,
                 clear_cache: bool = False,
                 ) -> None:
    """
    Initialize Builder module.
    If use_cache is True then create cache directory if directory does not exist.

    :param path_to_cache: Absolute path to directory to contain cached objects
    :param use_cache: True to use cached objects (default: False)
    :param use_timing: True to show timing information (default: False)
    :param clear_cache: True to clear delete cache directory and create new directory
    :return: None
    """
    # Update globals
    globals().update(global_path_to_cache=path_to_cache)
    globals().update(global_use_cache=use_cache)
    globals().update(global_use_timing=use_timing)
    globals().update(global_clear_cache=clear_cache)

    if global_clear_cache:
        shutil.rmtree(global_path_to_cache, ignore_errors=True)

    if global_use_cache:
        if not os.path.exists(global_path_to_cache):
            os.makedirs(global_path_to_cache, exist_ok=True)

    logger.use_log_perf = global_use_timing


def build_dataprovider(filename: str, use_cache: Union[bool, None] = None) -> DataProvider:
    """
    Build DataProvider.
    If use_cache is True then create DataProvider normally, write pickled object to file and read pickled
    object from cache.

    :param filename: Target settings filename
    :param use_cache: True to use cached DataProvider object (default: False)
    :return: DataProvider-object
    """
    if use_cache is None:
        use_cache = global_use_cache

    dataprovider = None
    if use_cache:
        path_to_cached_dataprovider = os.path.join(global_path_to_cache, "dataprovider.pickle")

        if not os.path.exists(path_to_cached_dataprovider):
            try:
                dataprovider = DataProvider(filename)
            except Exception as ex:
                show_exception_errors(ex, "Following errors occurred when loading settings file:")
                print("Fatal error, stopping execution...")
                raise ex

            with open(path_to_cached_dataprovider, "wb") as fs:
                pickle.dump(dataprovider, fs, pickle.HIGHEST_PROTOCOL)

        with open(path_to_cached_dataprovider, "rb") as fs:
            dataprovider = pickle.load(fs)
    else:
        try:
            dataprovider = DataProvider(filename)
        except Exception as ex:
            show_exception_errors(ex, "Following errors occurred when loading settings file:")
            print("Fatal error, stopping execution...")
            raise ex

    return dataprovider


def build_datachecker(dataprovider: DataProvider = None, use_cache: Union[bool, None] = None) -> DataChecker:
    """
    Build DataChecker.

    :param dataprovider: DataProvider
    :param use_cache: True to use cached DataChecker from file (default: False)
    :return: DataChecker
    """
    if use_cache is None:
        use_cache = global_use_cache

    datachecker = None
    if use_cache:
        path_to_cached_datachecker = os.path.join(global_path_to_cache, "datachecker.pickle")
        if not os.path.exists(path_to_cached_datachecker):
            datachecker = DataChecker(dataprovider)

            with open(path_to_cached_datachecker, "wb") as fs:
                pickle.dump(datachecker, fs, pickle.HIGHEST_PROTOCOL)

        with open(path_to_cached_datachecker, "rb") as fs:
            datachecker = pickle.load(fs)
    else:
        datachecker = DataChecker(dataprovider)

    return datachecker


def build_and_solve_scenarios(datachecker: DataChecker = None, use_cache: Union[bool, None] = None) -> List[Scenario]:
    """
    Build and check errors in scenario data and solve scenarios.

    :param datachecker: DataChecker-object
    :param use_cache: True to use cached Scenario-objects (default: False)
    :return: List of solved Scenario-objects
    """
    if use_cache is None:
        use_cache = global_use_cache

    scenarios = []
    if use_cache:
        # Check for cached scenario files
        filenames_to_load = []
        scenario_prefix = "scenario"
        files_in_cache = os.listdir(global_path_to_cache)
        for file in files_in_cache:
            abs_path = os.path.join(global_path_to_cache, file)
            has_prefix = file.startswith(scenario_prefix)
            is_file = os.path.isfile(abs_path)
            if has_prefix and is_file:
                filenames_to_load.append(abs_path)

        if not filenames_to_load:
            # No existing scenario data, build scenarios to pickle to files
            try:
                scenarios = datachecker.build_scenarios()
            except Exception as ex:
                show_exception_errors(ex, "Following errors occurred when building scenarios:")
                print("Fatal error, stopping execution...")
                raise ex

            # Check for build scenario errors
            try:
                datachecker.check_for_errors()
            except Exception as ex:
                show_exception_errors(ex, "Following errors found when checking scenario errors:")
                print("Fatal error, stopping execution...")
                raise ex

            # Solve scenarios
            for scenario_index, scenario in enumerate(scenarios):
                # NOTE: Baseline scenario is always the first element in the list
                # and all the alternative scenarios (if any) are after that
                if scenario_index == 0:
                    # Process baseline scenario
                    baseline_flow_solver = FlowSolver(scenario=scenario)
                    baseline_flow_solver.solve_timesteps()
                    scenario.flow_solver = baseline_flow_solver
                else:
                    # Get and copy solved scenario data from baseline scenario flow solver
                    baseline_scenario_data = scenarios[0].flow_solver.get_solved_scenario_data()
                    scenario.copy_from_baseline_scenario_data(baseline_scenario_data)

                    # Solve this alternative scenario time steps
                    scenario_flow_solver = FlowSolver(scenario=scenario)
                    scenario_flow_solver.solve_timesteps()
                    scenario.flow_solver = scenario_flow_solver

            # Build MFA systems for the scenarios
            for scenario in scenarios:
                scenario.mfa_system = build_mfa_system_for_scenario(scenario)

            for scenario_index, scenario in enumerate(scenarios):
                filename = os.path.join(global_path_to_cache, "{}_{}.pickle".format(scenario_prefix, scenario_index))
                with open(filename, "wb") as fs:
                    pickle.dump(scenario, fs, pickle.HIGHEST_PROTOCOL)

            # Rescan cache for scenario files
            filenames_to_load = []
            files_in_cache = os.listdir(global_path_to_cache)
            for file in files_in_cache:
                if file.startswith(scenario_prefix) and os.path.isfile(file):
                    filenames_to_load.append(os.path.abspath(file))
            filenames_to_load.sort()

        # Load all scenario files
        for filename in filenames_to_load:
            with open(filename, "rb") as fs:
                scenario = pickle.load(fs)
                scenarios.append(scenario)

    else:
        try:
            scenarios = datachecker.build_scenarios()
        except Exception as ex:
            show_exception_errors(ex, "Following errors occurred when building scenarios:")
            print("Fatal error, stopping execution...")
            raise ex

        # Check for build scenario errors
        try:
            datachecker.check_for_errors()
        except Exception as ex:
            show_exception_errors(ex, "Following errors found when checking scenario errors:")
            print("Fatal error, stopping execution...")
            raise ex

        # Solve scenarios
        for scenario_index, scenario in enumerate(scenarios):
            # NOTE: Baseline scenario is always the first element in the list
            # and all the alternative scenarios (if any) are after that
            if scenario_index == 0:
                # Process baseline scenario
                baseline_flow_solver = FlowSolver(scenario=scenario)
                baseline_flow_solver.solve_timesteps()
                scenario.flow_solver = baseline_flow_solver
            else:
                # Get and copy solved scenario data from baseline scenario flow solver
                baseline_scenario_data = scenarios[0].flow_solver.get_solved_scenario_data()
                scenario.copy_from_baseline_scenario_data(baseline_scenario_data)

                # Solve this alternative scenario time steps
                scenario_flow_solver = FlowSolver(scenario=scenario, reset_evaluated_values=False)
                scenario_flow_solver.solve_timesteps()
                scenario.flow_solver = scenario_flow_solver

        # Build MFA systems for the scenarios
        for scenario in scenarios:
            scenario.mfa_system = build_mfa_system_for_scenario(scenario)

    return scenarios


def build_results(filename: str = None, path_to_output_dir: Union[str, None] = None):
    """
    Build and solve scenarios using the settings file.

    :param filename: Path to Excel settings file
    :param path_to_output_dir: If None then uses the path from settings file
    :return: Tuple (model parameters (dictionary), list of scenarios (List[Scenario), color definitions (dictionary))
    """
    # Build DataProvider
    log("Loading data from file '{}'...".format(filename), level="info")

    start_log_perf("Loaded DataProvider" if not global_use_cache else "Loaded DataProvider (cached)")
    log("Build DataProvider...")
    dataprovider: DataProvider = build_dataprovider(filename)
    stop_log_perf()

    # Model parameters is a Dictionary that contains loaded data from Excel sheet named "Settings"
    # and are used for running the FlowSolver and setting up ODYM
    model_params = dataprovider.get_model_params()

    # If parameter "path_to_output_dir" is set then replace the output from
    # defined in scenario file with the "path_to_output_dir"
    use_path_to_output_from_settings = True
    if path_to_output_dir:
        use_path_to_output_from_settings = False
        model_params[ParameterName.OutputPath] = path_to_output_dir

    show_model_parameters(model_params)

    # Setup output path
    # Convert output directory name to absolute path and update model parameter dictionary
    if use_path_to_output_from_settings:
        model_params[ParameterName.OutputPath] = os.path.abspath(
            os.path.join(os.getcwd(), model_params[ParameterName.OutputPath]))

    # **************************************************************
    # * Step 2: Check data integrity and build data for FlowSolver *
    # **************************************************************
    log("Build and check errors in data...")
    start_log_perf("Loaded DataChecker" if not global_use_cache else "Loaded DataChecker (cached)")
    datachecker: DataChecker = build_datachecker(dataprovider)
    stop_log_perf()

    # **************************************************
    # * Step 3: Build and solve flows in all scenarios *
    # **************************************************
    log("Solve scenarios...")
    start_log_perf("Solve scenarios" if not global_use_cache else "Loaded scenarios (cached)")
    scenarios: List[Scenario] = build_and_solve_scenarios(datachecker)
    stop_log_perf()

    # Transformation stage color definitions
    color_definitions = {color.name: color.value for color in dataprovider.get_color_definitions()}

    # Show perf logs if toggled on
    show_log_perf_summary("build_scenarios()")
    clear_log_perf()

    return model_params, scenarios, color_definitions




#!/usr/bin/env python3
"""
Script for testing package structure before building

Test function must be:
- prefixed with "test_"
- contain docstring (comment wrapped in 3 quotation marks)
- return two parameters: first is number of passed test and second is number of failed tests

"""
import sys
import logging
from typing import Tuple


class LogFormatter(logging.Formatter):
    """Custom formatting class with colorsfor logging-instance"""
    GREEN = "\033[0;32m"
    RED = "\033[0;31m"
    WHITE = "\033[0;37m"
    RESET = "\033[0m"

    FORMATS = {
        logging.INFO: f"{GREEN}[%(levelname)s]{RESET} %(message)s"
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


# Global logger instance
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
ch = logging.StreamHandler(sys.stdout)
ch.setLevel(logging.INFO)
ch.setFormatter(LogFormatter())
logger.addHandler(ch)


def test_core_imports() -> Tuple[int, int]:
    """
    Verify core module imports
    """
    passed = 0
    failed = 0
    logger.info("Verify importing as package")
    # TODO: Test importing aiphoria as package inside venv and using it
    return passed, failed


if __name__ == "__main__":
    """
    Run all callable test functions inside this script
    that are prefixed by "test_"
    """
    name_to_entry = {}
    module = sys.modules[__name__]
    attribute_names = dir(module)
    for name in attribute_names:
        if name.startswith("_"):
            continue

        if not name.startswith("test_"):
            continue

        attr = getattr(module, name)
        if not callable(attr):
            continue

        desc = attr.__doc__ if hasattr(attr, "__doc__") else None
        if desc:
            desc = desc.strip()
        name_to_entry[name] = {"function": attr, "name": name, "description": desc, "is_success": False, "errors": []}

    num_test_functions = len(name_to_entry)
    num_total_tests_passed = 0
    num_total_tests_failed = 0
    logger.info("Running {} test functions".format(num_test_functions))
    logger.info("")

    # Function name to status
    name_to_errors = {}
    for index, (func_name, entry) in enumerate(name_to_entry.items()):
        func = entry["function"]
        func_desc = entry["description"]
        func_result = entry["is_success"]
        func_errors = entry["errors"]

        if not func_desc:
            # No docstring for test, treat this as error
            func_errors.append("No docstring for test function \"{}\"".format(func_name))
            entry["is_success"] = False
            num_total_tests_failed += 1
            continue

        test_passed = 0
        test_failed = 0
        try:
            logger.info("Test function {}/{}: {}".format(index + 1, num_test_functions, func_desc))
            test_passed, test_failed = func()
            is_success = (test_passed == 0 and test_failed == 0) or (test_passed > 0 and test_failed == 0)

            # Assume test function was successful test_passed == 0 and test_failed == 0
            # This happens if test function does not return two parameters
            if is_success:
                entry["is_success"] = is_success
                test_passed = test_passed if test_passed > 0 else 1
            else:
                entry["is_success"] = is_success
                test_failed = test_failed if test_failed > 0 else 1
                entry["errors"].append("Test function failed")

        except Exception as ex:
            func_errors.append("Function not returning tuple (passed, failed)")
            entry["is_success"] = False
            if not test_failed:
                test_failed = 1

        num_total_tests_passed += test_passed
        num_total_tests_failed += test_failed
        logger.info("")

    num_total_tests = num_total_tests_passed + num_total_tests_failed
    logger.info("RESULT: Passed: {}, failed: {}, total tests: {}".format(
        num_total_tests_passed,
        num_total_tests_failed,
        num_total_tests))

    if num_total_tests_failed == 0:
        logger.info("All test functions passed ({}/{})".format(num_test_functions, num_test_functions))
    else:
        failed_tests = [entry for name, entry in name_to_entry.items() if not entry["is_success"]]
        logger.info("{}/{} test functions failed {}:".format(
            len(failed_tests),
            num_test_functions,
            "test" if num_test_functions == 1 else "tests"))

        for entry in failed_tests:
            logger.info("Function: {} (Description: {}): {}".format(entry["name"], entry["description"], entry["errors"][0]))



import sys
from enum import Enum
from typing import Tuple, List, Dict, Any
import numpy as np
from .flowsolver import FlowSolver
from .datastructures import Flow, Scenario, FlowModifier, Process
from .parameters import ParameterScenarioType
from .types import FunctionType
from .logger import log


class FlowErrorType(str, Enum):
    """
    Flow error type
    """
    Undefined: str = "none"
    NotEnoughTotalOutflows: str = "not_enough_total_outflows"
    NotEnoughOppositeFlowShares: str = "not_enough_opposite_flow_shares"
    NotEnoughSiblingFlowShares: str = "not_enough_sibling_flow_shares"


class FlowModifierSolver(object):

    class FlowChangeEntry(object):
        """
        Helper class for storing evaluated flow values
        """
        def __init__(self,
                     year: int = 0,
                     flow_id: str = None,
                     value: float = 0.0,
                     evaluated_share: float = 0.0,
                     evaluated_value: float = 0.0,
                     evaluated_offset: float = 0.0,
                     evaluated_share_offset: float = 0.0,
                     ):
            self._year = year
            self._flow_id = flow_id
            self._value = value
            self._evaluated_share = evaluated_share
            self._evaluated_value = evaluated_value
            self._evaluated_offset = evaluated_offset
            self._evaluated_share_offset = evaluated_share_offset

        def __str__(self) -> str:
            return "FlowChangeEntry: year={}, flow_id={}, value={}, evaluated_share={}, evaluated_value={}, " \
                   "evaluated_offset={}, evaluated_share_offset={}".format(
                self.year, self.flow_id, self.value, self.evaluated_share, self.evaluated_value, self.evaluated_offset,
                self.evaluated_share_offset)

        @property
        def year(self) -> int:
            return self._year

        @property
        def flow_id(self) -> str:
            return self._flow_id

        @property
        def value(self) -> float:
            return self._value

        @property
        def evaluated_share(self) -> float:
            return self._evaluated_share

        @property
        def evaluated_value(self) -> float:
            return self._evaluated_value

        @property
        def evaluated_offset(self) -> float:
            return self._evaluated_offset

        @property
        def evaluated_share_offset(self) -> float:
            return self._evaluated_share_offset

    class FlowErrorEntry(object):
        def __init__(self, year: int,
                     total_outflows: float,
                     required_outflows: float,
                     flow_modifier_index: int,
                     error_type: FlowErrorType = FlowErrorType.Undefined,
                     data: Dict[str, Any] = None
                     ):

            if data is None:
                data = {}

            self._year = year
            self._outflows_total = total_outflows
            self._outflows_required = required_outflows
            self._flow_modifier_index = flow_modifier_index
            self._error_type = error_type
            self._data = data

        @property
        def year(self) -> int:
            """
            Get year

            :return: Year (int)
            """
            return self._year

        @property
        def outflows_total(self) -> float:
            """
            Get total outflows

            :return: Total outflows (float)
            """
            return self._outflows_total

        @property
        def outflows_required(self) -> float:
            """
            Get required outflows

            :return: Required outflows (float)
            """
            return self._outflows_required

        @property
        def outflows_missing(self) -> float:
            """
            Calculate missing outflows (total outflows - required outflows)

            :return: Missing outflows (float)
            """
            return self.outflows_total - self.outflows_required

        @property
        def flow_modifier_index(self) -> int:
            """
            Get flow modifier index causing the error

            :return: Flow modifier index (int)
            """
            return self._flow_modifier_index

        @property
        def error_type(self) -> FlowErrorType:
            """
            Get flow error type

            :return: FlowErrorType (Enum)
            """
            return self._error_type

        @error_type.setter
        def error_type(self, new_error_type) -> None:
            """
            Set flow error type

            :param new_error_type: New flow error type
            """
            self._error_type = new_error_type

        @property
        def data(self) -> Dict[str, Any]:
            """
            Get error data.

            :return: Dictionary [str, Any]
            """
            return self._data

        @data.setter
        def data(self, data: Dict[str, Any]) -> None:
            """
            Set error data.

            :param data: New data (Dictionary [str, Any])
            """
            self._data = data

    def __init__(self, flow_solver: FlowSolver, scenario_type: ParameterScenarioType):
        self._flow_solver: FlowSolver = flow_solver
        self._scenario_type: ParameterScenarioType = scenario_type

    def solve(self):
        if self._scenario_type == ParameterScenarioType.Unconstrained:
            log("Solving unconstrained scenario...")
            ok, errors = self._solve_unconstrained_scenario()
            if not ok:
                sys.stdout.flush()
                log("Errors in unconstrained scenario:", level="error")
                for error in errors:
                    print("\t" + error)
                log("Unconstrained scenario contained errors, stopping now...", level="error")

        if self._scenario_type == ParameterScenarioType.Constrained:
            log("Solving constrained scenario...")
            ok, errors = self._solve_constrained_scenario()
            if not ok:
                sys.stdout.flush()
                log("Errors in constrained scenario:", level="error")
                for error in errors:
                    print("\t" + error)
                log("Unconstrained scenario contained errors, stopping now...", level="error")
                sys.exit(-1)

        log("Scenario solving done")

    def _solve_unconstrained_scenario(self) -> Tuple[bool, List[str]]:
        # ***************************************************************************
        # * Solve unconstrained scenario                                            *
        # * Introduces virtual flows if detecting that processes do not have enough *
        # * outflows and does not stop execution                                    *
        # ***************************************************************************
        errors: List[str] = []
        flow_solver: FlowSolver = self._flow_solver
        scenario: Scenario = self._flow_solver.get_scenario()

        scenario_type = ParameterScenarioType.Unconstrained
        flow_solver._reset_evaluated_values = True

        # Evaluate new values for each flow modifier in the requested year range
        # and group flow modifiers by source process ID. This is needed when multiple
        # flow modifiers are affecting the same source process.
        source_process_id_to_flow_modifier_indices = {}
        flow_modifier_index_to_new_values = {}
        flow_modifier_index_to_new_offsets = {}
        flow_modifiers = scenario.scenario_definition.flow_modifiers
        for flow_modifier_index, flow_modifier in enumerate(flow_modifiers):
            new_values, new_offsets = self._calculate_new_flow_values(flow_modifier)
            flow_modifier_index_to_new_values[flow_modifier_index] = new_values
            flow_modifier_index_to_new_offsets[flow_modifier_index] = new_offsets
            source_process_id = flow_modifier.source_process_id
            if source_process_id not in source_process_id_to_flow_modifier_indices:
                source_process_id_to_flow_modifier_indices[source_process_id] = []
            source_process_id_to_flow_modifier_indices[source_process_id].append(flow_modifier_index)

        # Separate into entries that affect relative and absolute flows by
        # checking what type of flow (absolute/relative) flow_modifier is targeting.
        # This is needed because the FlowModifiers only affect the same type of flows
        # as the source-to-target flow is.
        has_errors = False
        for source_process_id, flow_modifier_indices in source_process_id_to_flow_modifier_indices.items():
            flow_modifier_indices_for_abs_flows = []
            flow_modifier_indices_for_rel_flows = []
            for flow_modifier_index in flow_modifier_indices:
                flow_modifier = flow_modifiers[flow_modifier_index]
                flow = flow_solver.get_flow(flow_modifier.target_flow_id, flow_modifier.start_year)
                if flow.is_unit_absolute_value:
                    flow_modifier_indices_for_abs_flows.append(flow_modifier_index)
                else:
                    flow_modifier_indices_for_rel_flows.append(flow_modifier_index)

            # Cache process total absolute and relative outflows for every year
            # before any changes applied. This is used when recalculating new flow share
            year_to_total_outflows_abs = {}
            year_to_total_outflows_rel = {}
            for year in scenario.scenario_data.years:
                # NOTE: Every year might not contain all process IDs to skip those years
                has_total_abs = False
                has_total_rel = False
                try:
                    year_to_total_outflows_abs[year] = flow_solver.get_process_outflows_total_abs(source_process_id, year)
                    has_total_abs = True
                except KeyError:
                    pass
                try:
                    year_to_total_outflows_rel[year] = flow_solver.get_process_outflows_total_rel(source_process_id, year)
                    has_total_rel = True
                except KeyError:
                    pass

                if not has_total_abs and not has_total_rel:
                    continue

            # Solve absolute flows and relative flows independently
            abs_flow_modifier_index_to_error_entry, abs_changeset = self._process_absolute_flows(
                source_process_id,
                flow_solver,
                flow_modifier_indices_for_abs_flows,
                flow_modifiers,
                flow_modifier_index_to_new_values,
                flow_modifier_index_to_new_offsets,
                scenario_type)

            rel_flow_modifier_index_to_error_entry, rel_changeset = self._process_relative_flows(
                source_process_id,
                flow_solver,
                flow_modifier_indices_for_rel_flows,
                flow_modifiers,
                flow_modifier_index_to_new_values,
                flow_modifier_index_to_new_offsets,
                scenario_type)

            # Check if target opposite flows or sibling flows have enough flows for the flow modifiers
            self._check_flow_modifier_changes(flow_solver,
                                              flow_modifier_indices,
                                              flow_modifiers,
                                              rel_changeset)

            # *************************************************************
            # * Apply changesets (order: absolute flows, relative flows)  *
            # * This order is needed because relative flow values depends *
            # * on the remaining process outflows after applying absolute *
            # * flows values                                              *
            # *************************************************************
            # Apply changes targeting absolute flows
            for flow_modifier_index, changeset in abs_changeset.items():
                flow_modifier = flow_modifiers[flow_modifier_index]
                entry: FlowModifierSolver.FlowChangeEntry
                for entry in changeset:
                    flow = flow_solver.get_flow(entry.flow_id, entry.year)
                    if flow.id == flow_modifier.target_flow_id:
                        # Apply calculated changes to source-to-target flow
                        # This is because that entry is always first in the list
                        # Overwrites the flow value
                        flow.value = entry.value
                        flow.evaluated_value = entry.evaluated_value
                        flow.evaluated_share = entry.evaluated_share
                    else:
                        # Apply calculated offset to evaluated value, these are all sibling flows
                        # or the target opposite flows
                        flow.value += entry.evaluated_offset
                        flow.evaluated_value += entry.evaluated_offset

            # Apply changes targeting relative flows
            for flow_modifier_index, changeset in rel_changeset.items():
                flow_modifier = flow_modifiers[flow_modifier_index]
                entry: FlowModifierSolver.FlowChangeEntry
                for entry in changeset:
                    flow = flow_solver.get_flow(entry.flow_id, entry.year)
                    if flow.id == flow_modifier.target_flow_id:
                        # Apply calculated changes to source-to-target flow
                        # This is because that entry is always first in the list
                        # Overwrites the flow share
                        flow.value = entry.value
                        flow.evaluated_value = entry.evaluated_value
                        flow.evaluated_share = entry.evaluated_share
                    else:
                        # Apply calculated offset to evaluated value, these are all sibling flows
                        # or the target opposite flows
                        total_outflows_rel = year_to_total_outflows_rel[entry.year]
                        new_evaluated_value = flow.evaluated_value + entry.evaluated_offset
                        new_value = new_evaluated_value / total_outflows_rel * 100.0
                        new_evaluated_share = new_value / 100.0

                        flow.value += entry.evaluated_share_offset
                        flow.evaluated_value = new_evaluated_value
                        flow.evaluated_share = new_evaluated_share

            if abs_flow_modifier_index_to_error_entry:
                # Errors in absolute flows: Unpack error entries and show errors but do not stop execution
                print("[Unconstrained scenario] Found issues in scenarios targeting absolute flows:")
                has_errors = True
                min_error_entry = None
                flow_modifier_index: int
                error_entry: FlowModifierSolver.FlowErrorEntry
                for flow_modifier_index, error_entry in abs_flow_modifier_index_to_error_entry.items():
                    # Gather only total outflow errors
                    if error_entry.error_type != FlowErrorType.NotEnoughTotalOutflows:
                        continue

                    if not min_error_entry:
                        min_error_entry = error_entry
                        continue

                    if error_entry.outflows_missing < min_error_entry.outflows_missing:
                        min_error_entry = error_entry

                if min_error_entry:
                    year = min_error_entry.year
                    total = min_error_entry.outflows_total
                    required = min_error_entry.outflows_required
                    missing = min_error_entry.outflows_missing
                    flow_modifier = flow_modifiers[min_error_entry.flow_modifier_index]

                    # TODO: Show target relative share that allows to scenario to work in
                    # TODO: error instead of absolute numbers
                    # TODO: Show easy-to-understand error message saying that
                    # TODO: a) increase/decrease the change in value to this amount to make this work
                    # TODO: b) this is the minimum/maximum target value that can be used here

                    s = "Process '{}'".format(source_process_id)
                    s += " "
                    s += "does not have enough outflows for absolute flows in year {}".format(year)
                    s += " "
                    s += "(total={}, required={}, missing={})".format(total, required, missing)
                    s += " "
                    s += "(row number {})".format(flow_modifier.row_number)
                    print("\tERROR: {}".format(s))

            if rel_flow_modifier_index_to_error_entry:
                # All flow modifiers in rel_flow_modifier_index_to_error_entry-map points to same source process ID
                print("[Unconstrained scenario] Found issues in scenarios targeting relative flows:")
                has_errors = True
                min_error_entry = None
                flow_modifier_index: int
                error_entry: FlowModifierSolver.FlowErrorEntry
                for flow_modifier_index, error_entry in rel_flow_modifier_index_to_error_entry.items():
                    # Gather only total outflow errors
                    if error_entry.error_type != FlowErrorType.NotEnoughTotalOutflows:
                        continue

                    if not min_error_entry:
                        min_error_entry = error_entry
                        continue

                    if error_entry.outflows_missing < min_error_entry.outflows_missing:
                        min_error_entry = error_entry

                if min_error_entry:
                    year = min_error_entry.year
                    total = min_error_entry.outflows_total
                    required = min_error_entry.outflows_required
                    missing = min_error_entry.outflows_missing
                    flow_modifier = flow_modifiers[min_error_entry.flow_modifier_index]

                    s = "Process '{}'".format(source_process_id)
                    s += " "
                    s += "does not have enough outflows for relative flows in year {}".format(year)
                    s += " "
                    s += "(total={}, required={}, missing={})".format(total, required, missing)
                    s += " "
                    s += "(row number {})".format(flow_modifier.row_number)
                    print("\tERROR: {}".format(s))

                for flow_modifier_index, error_entry in rel_flow_modifier_index_to_error_entry.items():
                    flow_modifier = flow_modifiers[flow_modifier_index]
                    if error_entry.error_type is not FlowErrorType.NotEnoughOppositeFlowShares:
                        continue

                    error_data = error_entry.data
                    year = error_data["year"]
                    if flow_modifier.use_change_in_value:
                        opposite_target_flow_ids = error_data["opposite_target_flow_ids"]

                        # required_flow_shares is flow share value between [0, 1]
                        # available_flow_shares is flow share value between [0, 1]
                        required_flow_shares = error_data["required_flow_shares"]
                        available_flow_shares = error_data["available_flow_shares"]

                        # Calculate maximum evaluated share for the flow modifier in range [0, 1]
                        # and then convert maximum evaluated share to maximum possible change in value
                        target_flow = flow_solver.get_flow(flow_modifier.target_flow_id, flow_modifier.start_year)
                        start_year_evaluated_share = target_flow.evaluated_share
                        max_evaluated_share = start_year_evaluated_share + available_flow_shares
                        max_change_in_value = ((max_evaluated_share / start_year_evaluated_share) - 1) * 100.0

                        # Round value down so there value shown here is less than the absolute maximum value
                        max_change_in_value -= 0.001
                        s = "Flow modifier in row {} targets opposite flows that do not have enough flow shares".format(
                            flow_modifier.row_number)
                        s += " "
                        s += "(required={}, available={}).".format(required_flow_shares * 100.0,
                                                                   available_flow_shares * 100.0)
                        s += " "
                        s += "Maximum available change in value is {:.3f}%".format(max_change_in_value)
                        print("\tERROR: {}".format(s))
                    else:
                        print("TODO: Implement showing for error when using target value")

            if has_errors:
                sys.exit(-1)
                pass

        # Recalculate process inflows/outflows
        self._recalculate_relative_flow_evaluated_shares(flow_solver)

        # Check if flow modifiers caused negative flows to target opposite flows
        errors += self._check_flow_modifier_results(flow_solver, flow_modifiers)

        # Clamp all flows to minimum of 0.0 to introduce virtual flows
        flow_solver.clamp_flow_values_below_zero()

        return not errors, errors

    def _solve_constrained_scenario(self) -> Tuple[bool, List[str]]:
        # ********************************************************************************
        # * Solve constrained scenario                                                   *
        # * Difference to unconstrained scenario is that no virtual flows are introduced *
        # * and execution stops if errors are found                                      *
        # ********************************************************************************
        errors: List[str] = []
        flow_solver: FlowSolver = self._flow_solver
        scenario: Scenario = self._flow_solver.get_scenario()

        scenario_type = ParameterScenarioType.Constrained
        flow_solver._reset_evaluated_values = True

        # Evaluate new values for each flow modifier in the requested year range
        # and group flow modifiers by source process ID. This is needed when multiple
        # flow modifiers are affecting the same source process.
        source_process_id_to_flow_modifier_indices = {}
        flow_modifier_index_to_new_values = {}
        flow_modifier_index_to_new_offsets = {}
        flow_modifiers = scenario.scenario_definition.flow_modifiers
        for flow_modifier_index, flow_modifier in enumerate(flow_modifiers):
            new_values, new_offsets = self._calculate_new_flow_values(flow_modifier)
            flow_modifier_index_to_new_values[flow_modifier_index] = new_values
            flow_modifier_index_to_new_offsets[flow_modifier_index] = new_offsets
            source_process_id = flow_modifier.source_process_id
            if source_process_id not in source_process_id_to_flow_modifier_indices:
                source_process_id_to_flow_modifier_indices[source_process_id] = []
            source_process_id_to_flow_modifier_indices[source_process_id].append(flow_modifier_index)

        # Separate into entries that affect relative and absolute flows by
        # checking what type of flow (absolute/relative) flow_modifier is targeting.
        # This is needed because the FlowModifiers only affect the same type of flows
        # as the source-to-target flow is.
        has_errors = False
        for source_process_id, flow_modifier_indices in source_process_id_to_flow_modifier_indices.items():
            flow_modifier_indices_for_abs_flows = []
            flow_modifier_indices_for_rel_flows = []
            for flow_modifier_index in flow_modifier_indices:
                flow_modifier = flow_modifiers[flow_modifier_index]
                flow = flow_solver.get_flow(flow_modifier.target_flow_id, flow_modifier.start_year)
                if flow.is_unit_absolute_value:
                    flow_modifier_indices_for_abs_flows.append(flow_modifier_index)
                else:
                    flow_modifier_indices_for_rel_flows.append(flow_modifier_index)

            # Cache process total absolute and relative outflows for every year
            # before any changes applied. This is used when recalculating new flow share
            year_to_total_outflows_abs = {}
            year_to_total_outflows_rel = {}
            for year in scenario.scenario_data.years:
                # NOTE: Every year might not contain all process IDs to skip those years
                has_total_abs = False
                has_total_rel = False
                try:
                    year_to_total_outflows_abs[year] = flow_solver.get_process_outflows_total_abs(source_process_id, year)
                    has_total_abs = True
                except KeyError:
                    pass
                try:
                    year_to_total_outflows_rel[year] = flow_solver.get_process_outflows_total_rel(source_process_id, year)
                    has_total_rel = True
                except KeyError:
                    pass

                if not has_total_abs and not has_total_rel:
                    continue

            # Solve absolute flows and relative flows independently
            abs_flow_modifier_index_to_error_entry, abs_changeset = self._process_absolute_flows(
                source_process_id,
                flow_solver,
                flow_modifier_indices_for_abs_flows,
                flow_modifiers,
                flow_modifier_index_to_new_values,
                flow_modifier_index_to_new_offsets,
                scenario_type)

            rel_flow_modifier_index_to_error_entry, rel_changeset = self._process_relative_flows(
                source_process_id,
                flow_solver,
                flow_modifier_indices_for_rel_flows,
                flow_modifiers,
                flow_modifier_index_to_new_values,
                flow_modifier_index_to_new_offsets,
                scenario_type)

            # *************************************************************
            # * Apply changesets (order: absolute flows, relative flows)  *
            # * This order is needed because relative flow values depends *
            # * on the remaining process outflows after applying absolute *
            # * flows values                                              *
            # *************************************************************
            # Apply changes targeting absolute flows
            for flow_modifier_index, changeset in abs_changeset.items():
                flow_modifier = flow_modifiers[flow_modifier_index]
                entry: FlowModifierSolver.FlowChangeEntry
                for entry in changeset:
                    flow = flow_solver.get_flow(entry.flow_id, entry.year)
                    if flow.id == flow_modifier.target_flow_id:
                        # Apply calculated changes to source-to-target flow
                        # This is because that entry is always first in the list
                        # Overwrites the flow value
                        flow.value = entry.value
                        flow.evaluated_value = entry.evaluated_value
                        flow.evaluated_share = entry.evaluated_share
                    else:
                        # Apply calculated offset to evaluated value, these are all sibling flows
                        # or the target opposite flows
                        flow.value += entry.evaluated_offset
                        flow.evaluated_value += entry.evaluated_offset

            # Apply changes targeting relative flows
            for flow_modifier_index, changeset in rel_changeset.items():
                flow_modifier = flow_modifiers[flow_modifier_index]
                entry: FlowModifierSolver.FlowChangeEntry
                for entry in changeset:
                    flow = flow_solver.get_flow(entry.flow_id, entry.year)
                    if flow.id == flow_modifier.target_flow_id:
                        # Apply calculated changes to source-to-target flow
                        # This is because that entry is always first in the list
                        # Overwrites the flow share
                        flow.value = entry.value
                        flow.evaluated_value = entry.evaluated_value
                        flow.evaluated_share = entry.evaluated_share
                    else:
                        # Apply calculated offset to evaluated value, these are all sibling flows
                        # or the target opposite flows
                        total_outflows_rel = year_to_total_outflows_rel[entry.year]
                        new_evaluated_value = flow.evaluated_value + entry.evaluated_offset
                        new_value = new_evaluated_value / total_outflows_rel * 100.0
                        new_evaluated_share = new_value / 100.0

                        flow.value += entry.evaluated_share_offset
                        flow.evaluated_value = new_evaluated_value
                        flow.evaluated_share = new_evaluated_share

            if abs_flow_modifier_index_to_error_entry:
                # Errors in absolute flows: Unpack error entries and show errors but do not stop execution
                print("[Constrained scenario] Found issues in scenarios targeting absolute flows:")
                has_errors = True
                min_error_entry = None
                flow_modifier_index: int
                error_entry: FlowModifierSolver.FlowErrorEntry
                for flow_modifier_index, error_entry in abs_flow_modifier_index_to_error_entry.items():
                    # Gather only total outflow errors
                    if not error_entry.error_type == FlowErrorType.NotEnoughTotalOutflows:
                        continue

                    if not min_error_entry:
                        min_error_entry = error_entry
                        continue

                    if error_entry.outflows_missing < min_error_entry.outflows_missing:
                        min_error_entry = error_entry

                if min_error_entry:
                    year = min_error_entry.year
                    total = min_error_entry.outflows_total
                    required = min_error_entry.outflows_required
                    missing = min_error_entry.outflows_missing
                    flow_modifier = flow_modifiers[min_error_entry.flow_modifier_index]

                    s = "Process '{}'".format(source_process_id)
                    s += " "
                    s += "does not have enough outflows for absolute flows in year {}".format(year)
                    s += " "
                    s += "(total={}, required={}, missing={})".format(total, required, missing)
                    s += " "
                    s += "(row number {})".format(flow_modifier.row_number)
                    print("\tERROR: {}".format(s))

            if rel_flow_modifier_index_to_error_entry:
                # All flow modifiers in rel_flow_modifier_index_to_error_entry-map points to same source process ID
                print("[Constrained scenario] Found issues in scenarios targeting relative flows:")
                has_errors = True
                min_error_entry = None
                flow_modifier_index: int
                error_entry: FlowModifierSolver.FlowErrorEntry
                for flow_modifier_index, error_entry in rel_flow_modifier_index_to_error_entry.items():
                    # Gather only total outflow errors
                    if not error_entry.error_type == FlowErrorType.NotEnoughTotalOutflows:
                        continue

                    if not min_error_entry:
                        min_error_entry = error_entry
                        continue

                    if error_entry.outflows_missing < min_error_entry.outflows_missing:
                        min_error_entry = error_entry

                if min_error_entry:
                    year = min_error_entry.year
                    total = min_error_entry.outflows_total
                    required = min_error_entry.outflows_required
                    missing = min_error_entry.outflows_missing
                    flow_modifier = flow_modifiers[min_error_entry.flow_modifier_index]

                    s = "Process '{}'".format(source_process_id)
                    s += " "
                    s += "does not have enough outflows for relative flows in year {}".format(year)
                    s += " "
                    s += "(total={}, required={}, missing={})".format(total, required, missing)
                    s += " "
                    s += "(row number {})".format(flow_modifier.row_number)
                    print("\tERROR: {}".format(s))

        if has_errors:
            # Stop execution of constrained solver
            log("Errors in Constrained scenario solver, stopping execution...", level="error")
            sys.exit(-1)

        # Recalculate process inflows/outflows
        self._recalculate_relative_flow_evaluated_shares(flow_solver)

        # Check if flow modifiers caused negative flows to target opposite flows
        errors += self._check_flow_modifier_results(flow_solver, flow_modifiers)

        return not errors, errors

    def _get_process_outflow_siblings(self,
                                      process_id: str = None,
                                      flow_id: str = None,
                                      year: int = -1,
                                      only_same_type: bool = False,
                                      excluded_flow_ids: List[str] = None
                                      ) -> List[Flow]:
        """
        Get all sibling outflows for process ID and outflow ID.
        If only_same_type is True then return only outflows that are same type as flow_id.
        NOTE: Target flow (flow_id) is not included in the list of siblings.

        :param flow_id: Target Flow ID (excluded from results)
        :param year: Target year
        :param only_same_type: True to return only same type sibling outflows as flow_id, False returns all siblings.
        :param excluded_flow_ids: List of Flow IDs to exclude from siblings (optional)
        :return: List of Flows (List[Flow])
        """

        if excluded_flow_ids is None:
            excluded_flow_ids = []
        unique_excluded_flow_ids = set(excluded_flow_ids)

        all_outflows = {flow.id: flow for flow in self._flow_solver.get_process_outflows(process_id, year)}
        target_flow = all_outflows[flow_id]
        sibling_outflows = []
        for outflow_id, outflow in all_outflows.items():
            if outflow_id == flow_id:
                continue

            if outflow_id in unique_excluded_flow_ids:
                continue

            if only_same_type:
                if outflow.is_unit_absolute_value == target_flow.is_unit_absolute_value:
                    sibling_outflows.append(outflow)
            else:
                sibling_outflows.append(outflow)
        return sibling_outflows

    def _process_absolute_flows(self,
                                source_process_id: str,
                                flow_solver: FlowSolver,
                                flow_modifier_indices: List[int],
                                flow_modifiers: List[FlowModifier],
                                flow_modifier_index_to_new_values: Dict[int, List[float]],
                                flow_modifier_index_to_new_offsets: Dict[int, List[float]],
                                scenario_type: ParameterScenarioType
                                ) -> Tuple[Dict[int, FlowErrorEntry], Dict[int, List[FlowChangeEntry]]]:
        """
        Process absolute flows for flow modifier.

        :param source_process_id: Source Process ID
        :param flow_solver: FlowSolver
        :param flow_modifier_indices: List of flow modifier indices that affect absolute flows
        :param flow_modifiers: List of FlowModifiers
        :param flow_modifier_index_to_new_values: Mapping of flow modifier index to list of new values
        :param flow_modifier_index_to_new_offsets: Mapping of flow modifier index to list of offset values
        :return: Tuple (Dictionary (flow modifier index to FlowErrorEntry), Dictionary (flow modifier index to changeset)
        """
        # Flow modifier index to list of error entries
        flow_modifier_index_to_error_entry = {}

        # Flow modifier index to list of FlowChangeEntry-objects
        flow_modifier_index_to_changeset = {}

        # Flow modifier index to list of evaluated flow offset values
        flow_modifier_index_to_new_values_offset = {}

        # Flow modifier index to list of evaluated flow values
        flow_modifier_index_to_new_values_actual = {}

        # Year to total outflows required
        year_to_total_outflows_required = {}

        # Build yearly required outflows mapping
        for flow_modifier_index in flow_modifier_indices:
            flow_modifier = flow_modifiers[flow_modifier_index]
            new_values_offset = flow_modifier_index_to_new_offsets[flow_modifier_index]
            new_values = flow_modifier_index_to_new_values[flow_modifier_index]
            flow_modifier_index_to_new_values_offset[flow_modifier_index] = new_values_offset
            flow_modifier_index_to_new_values_actual[flow_modifier_index] = new_values
            for year_index, year in enumerate(flow_modifier.get_year_range()):
                if year not in year_to_total_outflows_required:
                    year_to_total_outflows_required[year] = 0.0
                year_to_total_outflows_required[year] += new_values[year_index]

        # Store year to source process total absolute outflows before applying changes
        year_to_process_total_outflows = {}
        for flow_modifier_index in flow_modifier_index_to_new_values_actual:
            flow_modifier = flow_modifiers[flow_modifier_index]
            for year in flow_modifier.get_year_range():
                # Update the yearly total absolute outflows only once because it stays the
                # same for all flow modifiers
                if year in year_to_process_total_outflows:
                    continue

                year_to_process_total_outflows[year] = flow_solver.get_process_outflows_total_abs(source_process_id, year)

            # Check if there is enough total outflows from the source process to fulfill the flow modifier requirements
            # before applying the changes
            year_to_total_outflows_available = {}
            for year, total_outflows in year_to_process_total_outflows.items():
                total_outflows_required = year_to_total_outflows_required[year]
                total_outflows_available = total_outflows - total_outflows_required
                year_to_total_outflows_available[year] = total_outflows_available

            # Find entry with minimum value in list
            year_to_value = {year: value for year, value in year_to_total_outflows_available.items() if value < 0.0}
            if year_to_value:
                entry = min(year_to_value.items(), key=lambda x: x[1])
                year, value = entry

                # Create new error entry
                outflows_total = year_to_process_total_outflows[year]
                outflows_required = year_to_total_outflows_required[year]
                new_error_entry = FlowModifierSolver.FlowErrorEntry(year,
                                                                    outflows_total,
                                                                    outflows_required,
                                                                    flow_modifier_index,
                                                                    FlowErrorType.NotEnoughTotalOutflows)

                flow_modifier_index_to_error_entry[flow_modifier_index] = new_error_entry

        # Exit early if there is errors
        if scenario_type == ParameterScenarioType.Constrained and flow_modifier_index_to_error_entry:
            return flow_modifier_index_to_error_entry, flow_modifier_index_to_changeset

        # Apply changes to source to target flows
        year_to_total_outflows_required = {}
        for flow_modifier_index in flow_modifier_indices:
            flow_modifier = flow_modifiers[flow_modifier_index]
            new_values_offset = flow_modifier_index_to_new_offsets[flow_modifier_index]
            new_values = flow_modifier_index_to_new_values_actual[flow_modifier_index]
            for year_index, year in enumerate(flow_modifier.get_year_range()):
                target_flow_id = flow_modifier.target_flow_id
                target_flow = flow_solver.get_flow(target_flow_id, year)

                value_offset = new_values_offset[year_index]
                value_actual = new_values[year_index]
                new_value = value_actual
                new_evaluated_share = 1.0
                new_evaluated_value = new_value * new_evaluated_share
                new_evaluated_offset = value_offset

                # Build evaluated offset mapping
                new_entry = FlowModifierSolver.FlowChangeEntry(year,
                                                               target_flow_id,
                                                               new_value,
                                                               new_evaluated_share,
                                                               new_evaluated_value,
                                                               new_evaluated_offset)

                if flow_modifier_index not in flow_modifier_index_to_changeset:
                    flow_modifier_index_to_changeset[flow_modifier_index] = []
                flow_modifier_index_to_changeset[flow_modifier_index].append(new_entry)

                # Recalculate year_to_total_outflows_required
                if year not in year_to_total_outflows_required:
                    year_to_total_outflows_required[year] = 0.0
                year_to_total_outflows_required[year] += target_flow.evaluated_value

        # Get list of all unique flow IDs used in all flow modifiers, these flows should be excluded from sibling flows
        excluded_flow_ids = set()
        for flow_modifier_index in flow_modifier_indices:
            flow_modifier = flow_modifiers[flow_modifier_index]

            # Ignore source to target flow ID
            source_to_target_flow_id = flow_modifier.target_flow_id
            excluded_flow_ids.add(source_to_target_flow_id)

            # Ignore all opposite target flow IDs
            for flow_id in flow_modifier.opposite_target_process_ids:
                excluded_flow_ids.add(flow_id)

        # Convert unique list of excluded flow IDs back to list
        excluded_flow_ids = list(excluded_flow_ids)

        # Apply flow modifiers to opposite targets or to all same type sibling flows
        for flow_modifier_index in flow_modifier_indices:
            flow_modifier = flow_modifiers[flow_modifier_index]
            year_range = flow_modifier.get_year_range()

            # NOTE: Skip applying changes to target flows (either siblings or target opposite flows) if set
            if not flow_modifier.apply_to_targets:
                continue

            # Flow value offset from first year flow value
            new_values_offset = flow_modifier_index_to_new_values_offset[flow_modifier_index]
            if flow_modifier.has_opposite_targets:
                for year_index, year in enumerate(year_range):
                    value_offset = new_values_offset[year_index]

                    # Calculate opposite flow share of the total opposite evaluated value
                    # Applying this factor allows the opposite flows to have different flow shares
                    total_opposite_flow_value = 0.0
                    for opposite_target_process_id in flow_modifier.opposite_target_process_ids:
                        opposite_flow_id = Flow.make_flow_id(source_process_id, opposite_target_process_id)
                        opposite_flow = flow_solver.get_flow(opposite_flow_id, year)
                        total_opposite_flow_value += opposite_flow.evaluated_value

                    for opposite_target_process_id in flow_modifier.opposite_target_process_ids:
                        opposite_flow_id = Flow.make_flow_id(source_process_id, opposite_target_process_id)
                        opposite_flow = flow_solver.get_flow(opposite_flow_id, year)
                        opposite_flow_share = opposite_flow.evaluated_value / total_opposite_flow_value

                        # Calculate changes, create new FlowChangeEntry and append it to changeset
                        new_value = (opposite_flow.evaluated_value - value_offset) * opposite_flow_share
                        new_evaluated_share = 1.0
                        new_evaluated_value = new_value
                        new_evaluated_offset = -value_offset * opposite_flow_share
                        new_entry = FlowModifierSolver.FlowChangeEntry(year,
                                                                       opposite_flow_id,
                                                                       new_value,
                                                                       new_evaluated_share,
                                                                       new_evaluated_value,
                                                                       new_evaluated_offset)

                        flow_modifier_index_to_changeset[flow_modifier_index].append(new_entry)

            else:
                # *************************************************************************
                # * Apply changes to proportionally to all siblings outflows of same type *
                # *************************************************************************
                for year_index, year in enumerate(year_range):
                    # Get total absolute outflows
                    total_outflows_abs = year_to_process_total_outflows[year]
                    total_outflows_required = year_to_total_outflows_required[year]
                    total_outflows_available = total_outflows_abs - total_outflows_required
                    value_offset = new_values_offset[year_index]

                    # Get all same type sibling outflows (= outflows that start from same source process
                    # and are same type as the source to target flow)
                    sibling_outflows = self._get_process_outflow_siblings(source_process_id,
                                                                          flow_modifier.target_flow_id,
                                                                          year,
                                                                          only_same_type=True,
                                                                          excluded_flow_ids=excluded_flow_ids)

                    # Get total sibling outflows, used to check if there is enough outflows
                    # to fulfill the flow_modifier request
                    total_sibling_outflows = np.sum([flow.evaluated_value for flow in sibling_outflows])

                    # Calculate new sibling values and update sibling flows
                    for flow in sibling_outflows:
                        # Calculate changes, create new FlowChangeEntry and append it to changeset
                        sibling_flow_id = flow.id
                        new_value = 0.0
                        new_evaluated_share = 1.0
                        new_evaluated_value = 0.0
                        new_evaluated_offset = 0.0
                        if total_sibling_outflows > 0.0:
                            sibling_share = flow.evaluated_value / total_sibling_outflows
                            sibling_offset = -value_offset
                            new_value = (total_outflows_available * sibling_share) + sibling_offset * sibling_share
                            new_evaluated_share = 1.0
                            new_evaluated_value = new_value
                            new_evaluated_offset = sibling_offset * sibling_share

                        new_entry = FlowModifierSolver.FlowChangeEntry(year,
                                                                       sibling_flow_id,
                                                                       new_value,
                                                                       new_evaluated_share,
                                                                       new_evaluated_value,
                                                                       new_evaluated_offset)

                        flow_modifier_index_to_changeset[flow_modifier_index].append(new_entry)

        return flow_modifier_index_to_error_entry, flow_modifier_index_to_changeset

    def _process_relative_flows(self,
                                source_process_id: str,
                                flow_solver: FlowSolver,
                                flow_modifier_indices: List[int],
                                flow_modifiers: List[FlowModifier],
                                flow_modifier_index_to_new_values: Dict[int, List[float]],
                                flow_modifier_index_to_new_offsets: Dict[int, List[float]],
                                scenario_type: ParameterScenarioType
                                ) -> Tuple[Dict[int, FlowErrorEntry], Dict[int, List[FlowChangeEntry]]]:
        """
        Process relative flows for flow modifier.

        :param source_process_id: Source Process ID
        :param flow_solver: FlowSolver
        :param flow_modifier_indices: List of flow modifier indices that affect relative flows
        :param flow_modifiers: List of FlowModifiers
        :param flow_modifier_index_to_new_values: Mapping of flow modifier index to list of new values
        :param flow_modifier_index_to_new_offsets: Mapping of flow modifier index to list of offset values
        :return: Tuple (Dictionary (flow modifier index to FlowErrorEntry), Dictionary (flow modifier index to changeset))
        """
        # Flow modifier index to list of error entries
        flow_modifier_index_to_error_entry = {}

        # Flow modifier index to list of FlowChangeEntry-objects
        flow_modifier_index_to_changeset = {}

        # Flow modifier index to list of evaluated flow offset values
        flow_modifier_index_to_new_values_offset = {}

        # Flow modifier index to list of evaluated flow values
        flow_modifier_index_to_new_values_actual = {}

        # Year to total outflows required
        year_to_total_outflows_required = {}

        # Build yearly required outflows mapping
        for flow_modifier_index in flow_modifier_indices:
            flow_modifier = flow_modifiers[flow_modifier_index]
            new_values_offset = flow_modifier_index_to_new_offsets[flow_modifier_index]
            new_values = flow_modifier_index_to_new_values[flow_modifier_index]
            flow_modifier_index_to_new_values_offset[flow_modifier_index] = new_values_offset
            flow_modifier_index_to_new_values_actual[flow_modifier_index] = new_values
            for year_index, year in enumerate(flow_modifier.get_year_range()):
                total_outflows_rel = flow_solver.get_process_outflows_total_rel(flow_modifier.source_process_id, year)
                evaluated_value = (new_values[year_index] / 100.0) * total_outflows_rel
                if year not in year_to_total_outflows_required:
                    year_to_total_outflows_required[year] = 0.0
                year_to_total_outflows_required[year] += evaluated_value

        # Store year to source process total relative outflows before applying changes
        year_to_process_total_outflows = {}
        for flow_modifier_index in flow_modifier_indices:
            flow_modifier = flow_modifiers[flow_modifier_index]
            for year in flow_modifier.get_year_range():
                # Update the yearly total relative outflows only once because it stays the
                # same for all flow modifiers
                if year in year_to_process_total_outflows:
                    continue

                year_to_process_total_outflows[year] = flow_solver.get_process_outflows_total_rel(source_process_id, year)

            # Check if there is enough total outflows from the source process to fulfill the flow modifier requirements
            # before applying the changes
            year_to_total_outflows_available = {}
            for year, total_outflows in year_to_process_total_outflows.items():
                total_outflows_required = year_to_total_outflows_required[year]
                total_outflows_available = total_outflows - total_outflows_required
                year_to_total_outflows_available[year] = total_outflows_available

            # Find entry with minimum value in list
            year_to_value = {year: value for year, value in year_to_total_outflows_available.items() if value < 0.0}
            if year_to_value:
                entry = min(year_to_value.items(), key=lambda x: x[1])
                year, value = entry

                # Create new error entry
                outflows_total = year_to_process_total_outflows[year]
                outflows_required = year_to_total_outflows_required[year]
                new_error_entry = FlowModifierSolver.FlowErrorEntry(year,
                                                                    outflows_total,
                                                                    outflows_required,
                                                                    flow_modifier_index,
                                                                    FlowErrorType.NotEnoughTotalOutflows)

                flow_modifier_index_to_error_entry[flow_modifier_index] = new_error_entry

        # Check that there is enough flow share in target opposite flows
        for flow_modifier_index in flow_modifier_indices:
            flow_modifier = flow_modifiers[flow_modifier_index]
            source_process_id = flow_modifier.source_process_id
            new_values_offset = flow_modifier_index_to_new_values_offset[flow_modifier_index]
            new_values_actual = flow_modifier_index_to_new_values_actual[flow_modifier_index]

            # Check if target opposite flows have enough flow share to fulfill the target flow change
            if flow_modifier.has_opposite_targets:
                opposite_target_flow_ids = flow_modifier.get_opposite_target_flow_ids()

                year_to_required_flow_share = {}
                year_to_available_flow_shares = {}
                for year_index, year in enumerate(flow_modifier.get_year_range()):
                    available_opposite_flow_shares = 0.0
                    for opposite_flow_id in opposite_target_flow_ids:
                        opposite_flow = flow_solver.get_flow(opposite_flow_id, year)
                        available_opposite_flow_shares += opposite_flow.evaluated_share

                    # Map year to available flow shares and convert the required flow share to [0, 1] range
                    year_to_available_flow_shares[year] = available_opposite_flow_shares
                    year_to_required_flow_share[year] = (new_values_actual[year_index] - new_values_actual[0]) / 100.0

                for year_index, year in enumerate(flow_modifier.get_year_range()):
                    required_flow_shares = year_to_required_flow_share[year]
                    available_sibling_flow_shares = year_to_available_flow_shares[year]
                    if required_flow_shares > available_sibling_flow_shares:
                        outflows_total = year_to_process_total_outflows[year]
                        outflows_required = year_to_total_outflows_required[year]
                        data = {
                            "year": year,
                            "opposite_target_flow_ids": [opposite_target_flow_ids],
                            "required_flow_shares": required_flow_shares,
                            "available_flow_shares": available_sibling_flow_shares,
                        }

                        # Create new error entry
                        new_error_entry = FlowModifierSolver.FlowErrorEntry(year,
                                                                            outflows_total,
                                                                            outflows_required,
                                                                            flow_modifier_index,
                                                                            FlowErrorType.NotEnoughOppositeFlowShares,
                                                                            data,
                                                                            )

                        flow_modifier_index_to_error_entry[flow_modifier_index] = new_error_entry

            else:
                # Check if all sibling flows have enough flow share to fulfill the target flow change
                year_to_required_flow_share = {}
                year_to_available_flow_shares = {}
                for year_index, year in enumerate(flow_modifier.get_year_range()):
                    # Get available total sibling share for this year
                    sibling_flows = self._get_process_outflow_siblings(flow_modifier.source_process_id,
                                                                       flow_modifier.target_flow_id,
                                                                       year,
                                                                       only_same_type=True)

                    available_sibling_flow_shares = 0.0
                    for flow in sibling_flows:
                        available_sibling_flow_shares += flow.evaluated_share

                    # Map year to available flow shares and convert the required flow share to [0, 1] range
                    year_to_available_flow_shares[year] = available_sibling_flow_shares
                    year_to_required_flow_share[year] = (new_values_actual[year_index] - new_values_actual[0]) / 100.0

                for year_index, year in enumerate(flow_modifier.get_year_range()):
                    required_flow_shares = year_to_required_flow_share[year]
                    available_flow_shares = year_to_available_flow_shares[year]
                    if required_flow_shares > available_flow_shares:
                        outflows_total = year_to_process_total_outflows[year]
                        outflows_required = year_to_total_outflows_required[year]
                        data = {
                            "year": year,
                            "required_flow_shares": required_flow_shares,
                            "available_flow_shares": available_flow_shares,
                        }

                        # Create new error entry
                        new_error_entry = FlowModifierSolver.FlowErrorEntry(year,
                                                                            outflows_total,
                                                                            outflows_required,
                                                                            flow_modifier_index,
                                                                            FlowErrorType.NotEnoughSiblingFlowShares,
                                                                            data,
                                                                            )

                        flow_modifier_index_to_error_entry[flow_modifier_index] = new_error_entry

        # Exit early if there is errors
        if scenario_type == ParameterScenarioType.Constrained and flow_modifier_index_to_error_entry:
            return flow_modifier_index_to_error_entry, flow_modifier_index_to_changeset

        # Apply changes to source to target flows
        year_to_total_outflows_required = {}
        for flow_modifier_index in flow_modifier_indices:
            flow_modifier = flow_modifiers[flow_modifier_index]
            new_values_offset = flow_modifier_index_to_new_offsets[flow_modifier_index]
            new_values = flow_modifier_index_to_new_values_actual[flow_modifier_index]
            for year_index, year in enumerate(flow_modifier.get_year_range()):
                target_flow_id = flow_modifier.target_flow_id
                target_flow = flow_solver.get_flow(target_flow_id, year)
                total_outflows_rel = year_to_process_total_outflows[year]

                value_offset = new_values_offset[year_index]
                value_actual = new_values[year_index]
                new_value = value_actual
                new_evaluated_share = new_value / 100.0
                new_evaluated_value = new_evaluated_share * total_outflows_rel
                new_evaluated_offset = value_offset

                # Build evaluated offset mapping
                new_entry = FlowModifierSolver.FlowChangeEntry(year,
                                                               target_flow_id,
                                                               new_value,
                                                               new_evaluated_share,
                                                               new_evaluated_value,
                                                               new_evaluated_offset)

                if flow_modifier_index not in flow_modifier_index_to_changeset:
                    flow_modifier_index_to_changeset[flow_modifier_index] = []
                flow_modifier_index_to_changeset[flow_modifier_index].append(new_entry)

                # Recalculate year_to_total_outflows_required
                if year not in year_to_total_outflows_required:
                    year_to_total_outflows_required[year] = 0.0
                year_to_total_outflows_required[year] += target_flow.evaluated_value

        # Get list of all unique flow IDs used in all flow modifiers, these flows should be excluded from sibling flows
        excluded_flow_ids = set()
        for flow_modifier_index in flow_modifier_indices:
            flow_modifier = flow_modifiers[flow_modifier_index]

            # Ignore source to target flow ID
            source_to_target_flow_id = flow_modifier.target_flow_id
            excluded_flow_ids.add(source_to_target_flow_id)

            # Ignore all opposite target flow IDs
            for flow_id in flow_modifier.opposite_target_process_ids:
                excluded_flow_ids.add(flow_id)

        # Convert unique list of excluded flow IDs back to list
        excluded_flow_ids = list(excluded_flow_ids)

        # Apply flow modifiers to opposite targets or to all same type sibling flows
        for flow_modifier_index in flow_modifier_indices:
            flow_modifier = flow_modifiers[flow_modifier_index]
            year_range = flow_modifier.get_year_range()

            # NOTE: Skip applying changes to target flows (either siblings or target opposite flows) if set
            if not flow_modifier.apply_to_targets:
                continue

            # Flow share offset from first year flow share
            new_values_actual = flow_modifier_index_to_new_values_actual[flow_modifier_index]
            new_values_offset = flow_modifier_index_to_new_values_offset[flow_modifier_index]
            if flow_modifier.has_opposite_targets:
                for year_index, year in enumerate(year_range):
                    value_actual = new_values_actual[year_index]
                    value_offset = new_values_offset[year_index]
                    total_outflows_rel = year_to_process_total_outflows[year]

                    # Calculate opposite flow share of the total opposite evaluated value
                    # Applying this factor allows the opposite flows to have different flow shares
                    total_opposite_flow_value = 0.0
                    for opposite_target_process_id in flow_modifier.opposite_target_process_ids:
                        opposite_flow_id = Flow.make_flow_id(source_process_id, opposite_target_process_id)
                        opposite_flow = flow_solver.get_flow(opposite_flow_id, year)
                        total_opposite_flow_value += opposite_flow.evaluated_value

                    for opposite_target_process_id in flow_modifier.opposite_target_process_ids:
                        opposite_flow_id = Flow.make_flow_id(source_process_id, opposite_target_process_id)
                        opposite_flow = flow_solver.get_flow(opposite_flow_id, year)
                        opposite_flow_share = opposite_flow.evaluated_value / total_opposite_flow_value

                        # Calculate changes, create new FlowChangeEntry and append it to changeset
                        new_value = (opposite_flow.evaluated_value - value_offset) * opposite_flow_share
                        new_evaluated_share = new_value / total_outflows_rel
                        new_evaluated_value = new_value
                        new_evaluated_offset = -value_offset * opposite_flow_share
                        new_evaluated_share_offset = -(new_values_actual[year_index] - new_values_actual[0]) * opposite_flow_share

                        new_entry = FlowModifierSolver.FlowChangeEntry(year,
                                                                       opposite_flow_id,
                                                                       new_value,
                                                                       new_evaluated_share,
                                                                       new_evaluated_value,
                                                                       new_evaluated_offset,
                                                                       new_evaluated_share_offset)

                        flow_modifier_index_to_changeset[flow_modifier_index].append(new_entry)

            else:
                # *************************************************************************
                # * Apply changes to proportionally to all siblings outflows of same type *
                # *************************************************************************
                for year_index, year in enumerate(year_range):
                    # Get total relative outflows for process
                    total_outflows_rel = year_to_process_total_outflows[year]
                    total_outflows_required = year_to_total_outflows_required[year]
                    total_outflows_available = total_outflows_rel - total_outflows_required
                    value_offset = new_values_offset[year_index]

                    # Get all same type sibling outflows (= outflows that start from same source process
                    # and are same type as the source to target flow)
                    sibling_outflows = self._get_process_outflow_siblings(source_process_id,
                                                                          flow_modifier.target_flow_id,
                                                                          year,
                                                                          only_same_type=True,
                                                                          excluded_flow_ids=excluded_flow_ids)

                    # Get total sibling outflows, used to check if there is enough outflows
                    # to fulfill the flow_modifier request
                    total_sibling_outflows = np.sum([flow.evaluated_value for flow in sibling_outflows])

                    # Calculate new sibling values and update sibling flows
                    for flow in sibling_outflows:
                        # Calculate changes, create new FlowChangeEntry and append it to changeset
                        sibling_flow_id = flow.id
                        new_value = 0.0
                        new_evaluated_share = 1.0
                        new_evaluated_value = 0.0
                        new_evaluated_offset = 0.0
                        new_evaluated_share_offset = 0.0
                        if total_sibling_outflows > 0.0:
                            sibling_share = flow.evaluated_value / total_sibling_outflows
                            sibling_offset = -value_offset
                            new_value = (total_outflows_available * sibling_share) + sibling_offset * sibling_share
                            new_evaluated_share = 1.0
                            new_evaluated_value = new_value
                            new_evaluated_offset = sibling_offset * sibling_share
                            new_evaluated_share_offset = -(new_values_actual[year_index] - new_values_actual[0]) * sibling_share

                            # Handle FunctionType.Constant differently for relative flows
                            # There is no change in offset because of the offset stays the same during the whole
                            # year range of the FlowModifier
                            if flow_modifier.function_type == FunctionType.Constant:
                                new_evaluated_share_offset = new_evaluated_offset

                        new_entry = FlowModifierSolver.FlowChangeEntry(year,
                                                                       sibling_flow_id,
                                                                       new_value,
                                                                       new_evaluated_share,
                                                                       new_evaluated_value,
                                                                       new_evaluated_offset,
                                                                       new_evaluated_share_offset)

                        flow_modifier_index_to_changeset[flow_modifier_index].append(new_entry)

        return flow_modifier_index_to_error_entry, flow_modifier_index_to_changeset

    def _calculate_new_flow_values(self, flow_modifier: FlowModifier) -> Tuple[List[float], List[float]]:
        """
        Calculate new flow values for flow modifier.
        If flow_modifier targets absolute flow, returns tuple of (evaluated values, evaluated offset values)
        If flow_modifier targets relative flow, returns tuple of (evaluated flow shares, evaluated offset values)

        Does not modify the target flow.

        :param flow_modifier: Target FlowModifier
        :return: Tuple (list of evaluated flow values, list of evaluated flow value offsets)
        """

        flow_solver: FlowSolver = self._flow_solver
        year_range = flow_modifier.get_year_range()
        source_to_target_flow_id = flow_modifier.target_flow_id

        # ******************************************
        # * Create offset values for flow modifier *
        # ******************************************
        # Absolute flows: new evaluated flow value, relative flows: new evaluated flow share (0 - 100 range)
        new_values = [0.0 for _ in year_range]

        # Evaluated value from evaluated base value, always evaluated flow value (not share)
        new_offsets = [0.0 for _ in year_range]

        # Get total outflows (absolute + relative) for first year
        first_year = year_range[0]
        total_outflows_abs = flow_solver.get_process_outflows_total_abs(flow_modifier.source_process_id, first_year)
        total_outflows_rel = flow_solver.get_process_outflows_total_rel(flow_modifier.source_process_id, first_year)
        total_outflows = total_outflows_abs + total_outflows_rel
        first_year_flow = self._flow_solver.get_flow(source_to_target_flow_id, first_year)

        if flow_modifier.function_type == FunctionType.Constant:
            # NOTE: Constant replaces the values during the year range
            new_values = [flow_modifier.target_value for _ in year_range]

        # Change in value (delta)
        if flow_modifier.use_change_in_value:
            value_start = 0.0
            if first_year_flow.is_unit_absolute_value:
                # Absolute flow, use flow evaluated value as value_start
                value_start = first_year_flow.evaluated_value
            else:
                # Relative flow, use flow share as value_start
                value_start = first_year_flow.evaluated_share * 100.0

            if flow_modifier.function_type == FunctionType.Linear:
                new_values = np.linspace(start=0, stop=flow_modifier.change_in_value, num=len(year_range))

            if flow_modifier.function_type == FunctionType.Exponential:
                # NOTE: Is this function working properly with target value?
                new_values = np.logspace(start=0, stop=1, num=len(year_range))

            if flow_modifier.function_type == FunctionType.Sigmoid:
                # NOTE: Is this function working properly with target value?
                new_values = np.linspace(start=-flow_modifier.change_in_value,
                                         stop=flow_modifier.change_in_value,
                                         num=len(year_range))

                new_values = flow_modifier.change_in_value / (1.0 + np.exp(-new_values))

        # Target value (current to target)
        if flow_modifier.use_target_value:
            value_start = 0.0
            if first_year_flow.is_unit_absolute_value:
                # Absolute flow, use flow evaluated value as value_start
                value_start = first_year_flow.evaluated_value
            else:
                # Relative flow, use flow share as value_start
                value_start = first_year_flow.evaluated_share * 100.0

            if flow_modifier.function_type == FunctionType.Linear:
                new_values = np.linspace(start=value_start, stop=flow_modifier.target_value, num=len(year_range))

            if flow_modifier.function_type == FunctionType.Exponential:
                new_values = np.logspace(start=0, stop=1, num=len(year_range))

            if flow_modifier.function_type == FunctionType.Sigmoid:
                new_values = np.linspace(start=-flow_modifier.change_in_value,
                                         stop=flow_modifier.change_in_value,
                                         num=len(year_range))

                new_values = flow_modifier.change_in_value / (1.0 + np.exp(-new_values))

        # *******************************************************************************
        # * Calculate target values for flow modifier from start year and offset values *
        # *******************************************************************************
        for year_index, year in enumerate(year_range):
            base_value = first_year_flow.value
            base_evaluated_value = first_year_flow.evaluated_value
            base_evaluated_share = first_year_flow.evaluated_share

            # Absolute flow
            if first_year_flow.is_unit_absolute_value:
                if flow_modifier.is_change_type_value:
                    # Change by absolute value, either delta change or move toward target value
                    if flow_modifier.use_change_in_value:
                        # Increase/decrease by absolute value
                        offset = new_values[year_index]
                        new_values[year_index] = base_evaluated_value + offset

                        # Calculate evaluated offset from base evaluated value
                        new_offsets[year_index] = offset

                    if flow_modifier.use_target_value:
                        # Move toward absolute target value each year
                        offset = new_values[year_index]
                        new_values[year_index] = offset

                        # Calculate evaluated offset from base evaluated value
                        new_offsets[year_index] = offset - base_evaluated_value

                if flow_modifier.is_change_type_proportional:
                    # Proportional/percentual change of value, use delta change only
                    offset = new_values[year_index]
                    new_values[year_index] = base_evaluated_value + base_evaluated_value * offset / 100.0

                    # Calculate evaluated offset from base evaluated value
                    new_offsets[year_index] = base_evaluated_value * offset / 100.0

            # Relative flow
            else:
                if flow_modifier.use_change_in_value:
                    offset = new_values[year_index]
                    new_values[year_index] = base_value + base_evaluated_share * offset

                    # Calculate evaluated offset from base evaluated value
                    new_offsets[year_index] = base_evaluated_value * offset / 100.0

                if flow_modifier.use_target_value:
                    offset = new_values[year_index]
                    new_values[year_index] = offset

                    # Calculate evaluated offset from new flow share
                    new_offset = offset - base_evaluated_share * 100
                    new_evaluated_offset = (new_offset / 100.0) * total_outflows
                    new_offsets[year_index] = new_evaluated_offset

        return new_values, new_offsets

    def _check_flow_modifier_results(self,
                                     flow_solver: FlowSolver = None,
                                     flow_modifiers: List[FlowModifier] = None) -> List[str]:
        """
        Check if applying flow modifiers caused negative flows in target opposite flows.

        :param flow_solver: Target FlowSolver
        :param flow_modifiers: List of FlowModifiers
        :return: List of errors (empty list == no errors)
        """
        errors = []

        if flow_modifiers is None:
            flow_modifiers = []

        if not flow_solver:
            raise Exception("Parameter flow_solver is None, check calling code")

        # Check that all flows that are affected by the flow modifiers have evaluated value >= 0.0
        # This could be caused by flow modifier that has opposite target flows that do not have enough flow
        # and it will cause negative flow
        affected_flow_id_to_flow_modifier_indices = {}
        flow_modifier_index_to_year_to_affected_flow_ids = {}
        for flow_modifier_index, flow_modifier in enumerate(flow_modifiers):
            if flow_modifier_index not in flow_modifier_index_to_year_to_affected_flow_ids:
                flow_modifier_index_to_year_to_affected_flow_ids[flow_modifier_index] = {}

            for year in flow_modifier.get_year_range():
                year_to_affected_flows_ids = flow_modifier_index_to_year_to_affected_flow_ids[flow_modifier_index]

                if year not in year_to_affected_flows_ids:
                    year_to_affected_flows_ids[year] = []
                affected_flow_ids = year_to_affected_flows_ids[year]

                if flow_modifier.has_opposite_targets:
                    # # Get list of all opposite flow IDs
                    # for target_process_id in flow_modifier.opposite_target_process_ids:
                    #     opposite_flow_id = Flow.make_flow_id(flow_modifier.source_process_id, target_process_id)
                    #     affected_flow_ids.append(opposite_flow_id)
                    affected_flow_ids += flow_modifier.get_opposite_target_flow_ids()
                else:
                    # Get list of all same type sibling flows and unpack as flow IDs
                    sibling_flows = self._get_process_outflow_siblings(flow_modifier.source_process_id,
                                                                       flow_modifier.target_flow_id,
                                                                       year,
                                                                       only_same_type=True)

                    affected_flow_ids += [flow.id for flow in sibling_flows]

                for flow_id in affected_flow_ids:
                    if flow_id not in affected_flow_id_to_flow_modifier_indices:
                        affected_flow_id_to_flow_modifier_indices[flow_id] = set()
                    affected_flow_id_to_flow_modifier_indices[flow_id].add(flow_modifier_index)

        # Check if any affected flows is < 0.0 and find year with the smallest flow value
        for affected_flow_id, flow_modifier_indices in affected_flow_id_to_flow_modifier_indices.items():
            for flow_modifier_index in flow_modifier_indices:
                flow_modifier = flow_modifiers[flow_modifier_index]
                years = flow_modifier.get_year_range()
                year_to_evaluated_value = {}
                for year in years:
                    # NOTE: Baseline might have some processes that do not exist in the scenarios
                    # These are mostly virtual processes.
                    if not flow_solver.has_flow(affected_flow_id, year):
                        continue

                    flow = flow_solver.get_flow(affected_flow_id, year)
                    year_to_evaluated_value[year] = flow.evaluated_value

                # Find any negative flows
                negative_flows = [[k, v] for k, v in year_to_evaluated_value.items() if v < 0.0]
                if not negative_flows:
                    continue

                # Find entry with the smallest value
                min_year_entry = min(negative_flows, key=lambda x: x[1])
                s = "Flow modifier in row {} targets opposite flows that do not have enough flows. ".format(
                    flow_modifier.row_number)
                s += "This caused negative flow (evaluated value={}) for flow '{}' in year {}".format(
                    min_year_entry[1], affected_flow_id, min_year_entry[0])
                errors.append(s)

        # Using flow modifier solver with apply to targets = False can
        # make total relative outflows over 100%
        process_id_to_errors = {}
        total_rel_outflow_tolerance = 0.01
        for flow_modifier_index, flow_modifier in enumerate(flow_modifiers):
            source_process_id = flow_modifier.source_process_id
            for year in flow_modifier.get_year_range():
                outflows = flow_solver.get_process_outflows(source_process_id, year)
                total_outflows_rel = 0.0
                for flow in outflows:
                    if not flow.is_unit_absolute_value:
                        total_outflows_rel += flow.evaluated_share

                if total_outflows_rel > (1.0 + total_rel_outflow_tolerance):
                    s = "Flow modifier in row {} targeting flow {} causes the total relative outflows"
                    s += " of source process '{}' to become over 100% in year {} (evaluated share = {:.3f}%)"
                    s = s.format(flow_modifier.row_number,
                                 flow_modifier.target_flow_id,
                                 flow_modifier.source_process_id,
                                 year,
                                 (total_outflows_rel * 100.0)
                                 )
                    errors.append(s)

        return errors

    def _check_flow_modifier_changes(self,
                                     flow_solver: FlowSolver,
                                     flow_modifier_indices: List[int],
                                     flow_modifiers: List[FlowModifier],
                                     flow_change_entries: Dict[int, List[FlowChangeEntry]]):
        """
        Check if flow modifier changes can be done.
        Checks that:
        - target opposite flows do have enough share

        :param flow_solver: Target FlowSolver
        :param flow_modifier_indices: List of flow modifier indices
        :param flow_modifiers: List of all FlowModifiers
        :param flow_change_entries: Flow modification changeset
        """

        # Get available flow share for every flow modifier source process
        source_process_id_to_year_to_flow_id_to_available_share = {}
        for flow_modifier_index in flow_modifier_indices:
            flow_modifier = flow_modifiers[flow_modifier_index]
            source_process_id = flow_modifier.source_process_id

            if source_process_id not in source_process_id_to_year_to_flow_id_to_available_share:
                source_process_id_to_year_to_flow_id_to_available_share[source_process_id] = {}
            year_to_flow_id_to_available_share = source_process_id_to_year_to_flow_id_to_available_share[source_process_id]

            if flow_modifier.has_opposite_targets:
                for year in flow_modifier.get_year_range():
                    if year not in year_to_flow_id_to_available_share:
                        year_to_flow_id_to_available_share[year] = {}
                    flow_id_to_available_flow_share = year_to_flow_id_to_available_share[year]

                    opposite_flow_ids = flow_modifier.get_opposite_target_flow_ids()
                    for flow_id in opposite_flow_ids:
                        if flow_id in flow_id_to_available_flow_share:
                            continue

                        flow = flow_solver.get_flow(flow_id, year)
                        flow_id_to_available_flow_share[flow_id] = flow.evaluated_share

            else:
                for year in flow_modifier.get_year_range():
                    if year not in year_to_flow_id_to_available_share:
                        year_to_flow_id_to_available_share[year] = {}
                    flow_id_to_to_available_flow_share = year_to_flow_id_to_available_share[year]

                    sibling_flows = self._get_process_outflow_siblings(flow_modifier.source_process_id,
                                                                       flow_modifier.target_flow_id,
                                                                       year,
                                                                       only_same_type=True
                                                                       )

                    for flow in sibling_flows:
                        flow_id = flow.id
                        if flow_id in flow_id_to_to_available_flow_share:
                            continue

                        flow_id_to_to_available_flow_share[flow_id] = flow.evaluated_share

        # Map source process ID to year to available flow share
        source_process_id_to_year_to_available_flow_share = {}
        source = source_process_id_to_year_to_flow_id_to_available_share
        target = source_process_id_to_year_to_available_flow_share
        for source_process_id, year_to_flow_id_to_available_share in source.items():
            if source_process_id not in target:
                target[source_process_id] = {}
            target_year_to_available_flow_share = target[source_process_id]

            source_year_to_available_flow_share = source[source_process_id]
            for year, flow_id_to_available_share in source_year_to_available_flow_share.items():
                if year not in target_year_to_available_flow_share:
                    target_year_to_available_flow_share[year] = 0.0

                for flow_id, available_share in flow_id_to_available_share.items():
                    target_year_to_available_flow_share[year] += available_share

        # Reduce available flow shares from source_process_id_to_year_to_available_share by
        # applying all changeset entries
        source_process_id_to_year_to_required_flow_offset = {}
        for flow_modifier_index, changeset in flow_change_entries.items():
            flow_modifier = flow_modifiers[flow_modifier_index]
            source_process_id = flow_modifier.source_process_id

            if source_process_id not in source_process_id_to_year_to_required_flow_offset:
                source_process_id_to_year_to_required_flow_offset[source_process_id] = {}
            year_to_required_flow_offset = source_process_id_to_year_to_required_flow_offset[source_process_id]

            for entry in changeset:
                flow = flow_solver.get_flow(entry.flow_id, entry.year)
                flow_id = flow.id
                year = entry.year

                # Process only opposite and target flows
                if flow_id == flow_modifier.target_flow_id:
                    continue

                if year not in year_to_required_flow_offset:
                    year_to_required_flow_offset[year] = 0.0

                year_to_required_flow_offset[year] += (entry.evaluated_share_offset / 100.0)

        # Map source process ID to list of flow modifiers
        source_process_id_to_flow_modifier_indices = {}
        for flow_modifier_index in flow_modifier_indices:
            flow_modifier = flow_modifiers[flow_modifier_index]
            source_process_id = flow_modifier.source_process_id
            if source_process_id not in source_process_id_to_flow_modifier_indices:
                source_process_id_to_flow_modifier_indices[source_process_id] = []

            list_of_flow_modifier_indices = source_process_id_to_flow_modifier_indices[source_process_id]
            list_of_flow_modifier_indices.append(flow_modifier_index)

        # These map always have same process IDs and years
        source_process_id_to_error_entries = {}
        for source_process_id in source_process_id_to_year_to_available_flow_share:
            year_to_available_flow_share = source_process_id_to_year_to_available_flow_share[source_process_id]
            year_to_required_flow_offset = source_process_id_to_year_to_required_flow_offset[source_process_id]

            for year in year_to_available_flow_share.keys():
                available_flow_share = year_to_available_flow_share[year]
                required_flow_offset = year_to_required_flow_offset[year]

                if required_flow_offset > 0.0:
                    continue

                if available_flow_share < abs(required_flow_offset):
                    error_entry = [source_process_id, year, required_flow_offset, available_flow_share]
                    if source_process_id not in source_process_id_to_error_entries:
                        source_process_id_to_error_entries[source_process_id] = []
                    error_entries = source_process_id_to_error_entries[source_process_id]
                    error_entries.append(error_entry)

        # Check all error entries for source processes
        for source_process_id, error_entries in source_process_id_to_error_entries.items():
            local_flow_modifier_indices = source_process_id_to_flow_modifier_indices[source_process_id]

            # Find entry with most negative required flow offset, this is the largest value that is needed
            # to fulfill the flow modifier rule
            min_entry = min(error_entries, key=lambda x: x[2])
            min_source_process_id = min_entry[0]
            min_entry_year = min_entry[1]
            min_entry_required_flow_offset = min_entry[2]
            min_entry_available_flow_share = min_entry[3]

            # This is the maximum flow share that is available at the year when the flow share
            # became the most negative, use this value to calculate what is the maximum change in value
            # for every related flow modifier
            maximum_flow_share = min_entry_available_flow_share
            fmi_to_share_diff = {}
            fmi_to_start_value = {}
            for fmi in local_flow_modifier_indices:
                fm = flow_modifiers[fmi]
                start_year_flow = flow_solver.get_flow(fm.target_flow_id, fm.start_year)
                v0 = start_year_flow.evaluated_share
                v1 = start_year_flow.evaluated_share * (1.0 + (fm.change_in_value / 100.0))
                share_diff = v1 - v0
                fmi_to_share_diff[fmi] = share_diff
                fmi_to_start_value[fmi] = v0

            # Calculate what is the maximum change of value for each flow modifier that affects this flow
            # This is proportional value how much each flow modifier needs the flow:
            # - bigger growth in flow share will introduce larger factor from the maximum available flow share
            fmi_to_share = {}
            fmi_to_max_change_in_value = {}
            total_share_diff = sum([val for val in fmi_to_share_diff.values()])
            for fmi in local_flow_modifier_indices:
                fmi_to_share[fmi] = fmi_to_share_diff[fmi] / total_share_diff
                start_value = fmi_to_start_value[fmi]
                end_value = start_value + (maximum_flow_share * fmi_to_share[fmi])
                fmi_to_max_change_in_value[fmi] = ((end_value / start_value) - 1.0) * 100.0

            print("Adjust the following flow modifiers that affect the source process {}".format(source_process_id))
            for fmi in local_flow_modifier_indices:
                flow_modifier = flow_modifiers[fmi]
                max_change_in_value = fmi_to_max_change_in_value[fmi] - 0.001
                print("- Maximum change in value for flow modifier in row {} is {:.3f} %".format(
                    flow_modifier.row_number, max_change_in_value))

    def _recalculate_relative_flow_evaluated_shares(self, flow_solver: FlowSolver):
        """
        Recalculates flow shares after applying flow modifiers.

        :param flow_solver: Target FlowSolver
        :param flow_modifiers: List of FlowModifers
        :return: None
        """
        # Recalculate all relative flow shares
        year_to_process_to_flows = flow_solver.get_year_to_process_to_flows()
        for year, process_to_flows in year_to_process_to_flows.items():
            for process, flows in process_to_flows.items():
                process_id = process.id
                if process.is_virtual:
                    continue

                if process.stock_lifetime > 0:
                    continue

                outflows = flow_solver.get_process_outflows(process_id, year)
                total_outflows_rel = flow_solver.get_process_outflows_total_rel(process_id, year)
                outflows_rel = [f for f in outflows if not f.is_unit_absolute_value and not f.is_virtual]

                if total_outflows_rel <= 0.0:
                    continue

                for flow in outflows_rel:
                    flow.evaluated_share = flow.evaluated_value / total_outflows_rel



# Parameters names are defined in this file
# and DataProvider file provides the default
# values for each parameter
from enum import Enum
from typing import Dict, Any


class ParameterName(str, Enum):
    """
    Parameter enumerations used in settings file
    """

    # ***********************
    # * Required parameters *
    # ***********************

    # Process related
    SheetNameProcesses: str = "sheet_name_processes"
    SkipNumRowsProcesses: str = "skip_num_rows_processes"
    IgnoreColumnsProcesses: str = "ignore_columns_processes"

    # Flow related
    SheetNameFlows: str = "sheet_name_flows"
    SkipNumRowsFlows: str = "skip_num_rows_flows"
    IgnoreColumnsFlows: str = "ignore_columns_flows"

    # Model parameters
    StartYear: str = "start_year"
    EndYear: str = "end_year"
    DetectYearRange: str = "detect_year_range"
    UseVirtualFlows: str = "use_virtual_flows"
    VirtualFlowsEpsilon: str = "virtual_flows_epsilon"

    # ***********************
    # * Optional parameters *
    # ***********************
    ConversionFactorCToCO2: str = "conversion_factor_c_to_co2"
    FillMissingAbsoluteFlows: str = "fill_missing_absolute_flows"
    FillMissingRelativeFlows: str = "fill_missing_relative_flows"
    FillMethod: str = "fill_method"

    # Scenarios related
    UseScenarios: str = "use_scenarios"
    SheetNameScenarios: str = "sheet_name_scenarios"
    SkipNumRowsScenarios: str = "skip_num_rows_scenarios"
    IgnoreColumnsScenarios: str = "ignore_columns_scenarios"
    ScenarioType: str = "scenario_type"

    # Sheet name for colors
    SheetNameColors: str = "sheet_name_colors"
    SkipNumRowsColors: str = "skip_num_rows_colors"
    IgnoreColumnsColors: str = "ignore_columns_colors"

    # Network graph
    CreateNetworkGraphs: str = "create_network_graphs"
    CreateSankeyCharts: str = "create_sankey_charts"

    # Output path
    OutputPath: str = "output_path"

    # Show plots
    ShowPlots: str = "show_plots"

    # Visualize inflows to process IDs
    VisualizeInflowsToProcesses: str = "visualize_inflows_to_processes"

    # Baseline unit
    # BaseLineValueName: Name of the baseline value e.g. "Solid wood equivalent"
    # BaselineUnitName: Name of the baseline unit e.g. Mm3 SWE
    BaselineValueName: str = "baseline_value_name"
    BaselineUnitName: str = "baseline_unit_name"

    # Flow prioritization
    PrioritizeTransformationStages: str = "prioritize_transformation_stages"
    PrioritizeLocations: str = "prioritize_locations"

    # Sheet name for process positions
    SheetNameProcessPositions: str = "sheet_name_process_positions"

    # Check errors
    CheckErrors: str = "check_errors"

    # Include scenario file metadata
    IncludeMetadata: str = "include_metadata"


class ParameterFillMethod(str, Enum):
    """
    Valid values for parameter FillMethod
    """

    Zeros: str = "Zeros"
    Previous: str = "Previous"
    Next: str = "Next"
    Interpolate: str = "Interpolate"


class ParameterScenarioType(str, Enum):
    """
    Valid values for parameter ScenarioType
    """

    Constrained: str = "Constrained"
    Unconstrained: str = "Unconstrained"


# Parameters used for Process/Flows/Stocks in settings file
class StockDistributionType(str, Enum):
    """
    Stock distribution types and decay functions
    """
    Fixed: str = "Fixed"
    Normal: str = "Normal"
    LogNormal: str = "LogNormal"
    FoldedNormal: str = "FoldedNormal"
    Weibull: str = "Weibull"
    Simple: str = "Simple"
    LandfillDecayWood: str = "LandfillDecayWood"
    LandfillDecayPaper: str = "LandfillDecayPaper"


class LandfillDecayParameter(str, Enum):
    """
    Parameter names for LandFillDecay* types
    """
    Condition: str = "condition"


# *********************************
# * Stock distribution parameters *
# *********************************

class StockDistributionParameter(str, Enum):
    """
    Stock distribution parameters.
    """
    StdDev: str = "stddev"
    Shape: str = "shape"
    Scale: str = "scale"
    Condition: str = "condition"


class StockDistributionParameterValueType(object):
    """
    Storage class for stock distribution parameters.
    Emulates Dictionary behaviour allowing to get value using
    the [] notation e.g.
        StockDistributionParameterValueType[StockDistributionParameter.StdDev]
        StockDistributionParameterValueType["stddev"]
    """
    parameter_to_value_type = {
        StockDistributionParameter.StdDev: float,
        StockDistributionParameter.Shape: float,
        StockDistributionParameter.Scale: float,
        StockDistributionParameter.Condition: str,
    }

    def __class_getitem__(cls, item):
        """
        Get value type for StockDistributionParameter

        :param item: Item (str or StockDistributionParameter)
        :return:
        """
        item_enum = None
        if isinstance(item, str):
            try:
                item_enum = StockDistributionParameter(item)
            except ValueError:
                pass

        if item_enum is None:
            #raise Exception("{} is not valid StockDistributionParameter".format(item_enum))
            pass

        value_type = cls.parameter_to_value_type.get(item_enum, None)
        if value_type is None:
            #raise Exception("No value type for {}!".format(item_enum))
            pass

        return value_type


class RequiredStockDistributionParameters(object):
    """
    Storage class for required stock distribution parameters.
    Emulates Dictionary behaviour allowing to get value using
    the [] notation e.g.
        RequiredStockDistributionParameters[StockDistributionType.Fixed]
        RequiredStockDistributionParameters["Fixed"]
    """

    # NOTE: value must be dictionary
    # NOTE: Key = StockDistributionType, Value: Dictionary (str -> required type)
    stock_distribution_to_required_params = {
        StockDistributionType.Fixed: {},
        StockDistributionType.Normal: {"stddev": float},
        StockDistributionType.LogNormal: {"stddev": float},
        StockDistributionType.FoldedNormal: {"stddev": float},
        StockDistributionType.Weibull: {"shape": float, "scale": float},
        StockDistributionType.Simple: {},
        StockDistributionType.LandfillDecayWood: {"condition": str},
        StockDistributionType.LandfillDecayPaper: {"condition": str},
    }

    def __class_getitem__(cls, item) -> Dict[str, Any]:
        """
        Get required parameters for StockDistributionType.
        Parameter 'item' can be either string or StockDistributionType-instance.

        :param item: String or instance of StockDistributionType
        :return: Dictionary of required parameters (key = parameter name, value = type of required parameter e.g. float)
        """

        # Convert string to StockDistributionType Enum
        item_enum = None
        if isinstance(item, str):
            try:
                item_enum = StockDistributionType(item)
            except ValueError:
                pass

        if item_enum is None:
            # raise Exception("{} is not in RequiredStockDistributionParameters".format(item_name))
            pass

        params = cls.stock_distribution_to_required_params.get(item_enum, None)
        if params is None:
            # raise Exception("{} is not valid StockDistributionType!".format(item_enum))
            pass

        return params


class AllowedStockDistributionParameterValues(object):
    """
    Storage class for required stock distribution parameter values.
    Emulates Dictionary behaviour allowing to get value using
    the [] notation e.g.
        AllowedStockDistributionParameterValues[StockDistributionType.Fixed]
        AllowedStockDistributionParameterValues["Fixed"]
    """

    # NOTE: value must be dictionary
    # NOTE: Key = StockDistributionType, Value: Dictionary (str -> required type)
    stock_distribution_param_to_allowed_values = {
        StockDistributionParameter.Condition: {"Dry": str, "Wet": str, "Managed": str}
    }

    def __class_getitem__(cls, item) -> Dict[str, Any]:
        """
        Get allowed parameters for StockDistributionParameter.
        Parameter 'item' can be either string or StockDistributionParameter-instance.

        :param item: String or instance of StockDistributionParameter
        :return: Dictionary of allowed entries (key: StockDistributionParameter-instance, value: type)
        """

        # Convert string to StockDistributionParameter Enum
        item_enum = None
        if isinstance(item, str):
            try:
                item_enum = StockDistributionParameter(item)
            except ValueError:
                pass

        # If item_enum is not found in list of allowed values then return empty Dictionary
        params = cls.stock_distribution_param_to_allowed_values.get(item_enum, None)
        if params is None:
            params = {}

        return params


__version__ = "0.0.1"

# Import aiphoria core files, classes and functions
from . import core
from .core.builder import (
    init_builder,
    build_results,
    build_dataprovider,
    build_datachecker,
)

from .core.dataprovider import DataProvider
from .core.datachecker import DataChecker
from .core.datastructures import (
    Scenario,
    ScenarioData,
    Process,
    Flow,
    Stock,
    Indicator,
)

from .core.parameters import (
    ParameterName,
    ParameterFillMethod,
)

from .core.utils import (
    create_output_directory,
    setup_current_working_directory,
    set_output_directory,
    get_output_directory,
)

from .core.logger import log
from .runner import run_scenarios

__all__ = [
    "core",
    "init_builder",
    "build_results",
    "build_dataprovider",
    "build_datachecker",
    "DataProvider",
    "DataChecker",
    "Scenario",
    "Process",
    "Flow",
    "Stock",
    "Indicator",
    "ParameterName",
    "ParameterFillMethod",
    "create_output_directory",
    "setup_current_working_directory",
    "log",
    "set_output_directory",
    "get_output_directory",
    "run_scenarios",
]
